# this code is adopted from https://github.com/HySonLab/pandemic_tgnn/blob/main/code/models_multiresolution.py

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GCNConv

from .base import BaseModel


class MPNN_Encoder(nn.Module):
    def __init__(self, nfeat, nhid, nout, n_nodes, dropout):
        super(MPNN_Encoder, self).__init__()
        self.nhid = nhid
        self.n_nodes = n_nodes

        self.conv1 = GCNConv(nfeat, nhid)
        self.conv2 = GCNConv(nhid, nhid)
        self.bn1 = nn.BatchNorm1d(nhid)
        self.bn2 = nn.BatchNorm1d(nhid)

        self.fc1 = nn.Linear(nfeat+2*nhid, nhid )
        self.fc2 = nn.Linear(nhid, nout)

        self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU()

    def forward(self, adj, x):
        
        lst = list()
        weight = adj.to_sparse().values()
        adj = adj.to_sparse().indices()
        lst.append(x)

        x = self.relu(self.conv1(x,adj,edge_weight=weight))
        x = self.bn1(x.permute((0,2,1))).permute((0,2,1))
        x = self.dropout(x)
        lst.append(x)

        x = self.relu(self.conv2(x, adj,edge_weight=weight))
        x = self.bn2(x.permute((0,2,1))).permute((0,2,1))
        x = self.dropout(x)
        lst.append(x)

        x = torch.cat(lst, dim=2)
        x = self.relu(self.fc1(x.contiguous().view(-1, x.size(2))))
        x = self.dropout(x)
        x = self.relu(self.fc2(x))
        return x.view(-1, self.n_nodes, x.size(1))
    
    def reset_parameters(self):
        self.conv1.reset_parameters()
        self.conv2.reset_parameters()
        self.bn1.reset_parameters()
        self.bn2.reset_parameters()
        self.fc1.reset_parameters()
        self.fc2.reset_parameters()





class ATMGNN(BaseModel):
    """
    Attention-based Temporal Multiresolution Graph Neural Network (ATMGNN) 

    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features per node per timestep.
    num_timesteps_input : int
        Number of timesteps considered for each input sample (window size).
    num_timesteps_output : int
        Number of output timesteps to predict.
    nhid : int, optional
        Number of hidden units in the network and the output size of graph convolution layers. Default: 256.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    nhead : int, optional
        Number of heads in the multi-head attention mechanism. Default: 1.
    num_clusters : list, optional
        List of integers defining the number of clusters for each multiresolution layer. Default: [10, 5].
    use_norm : bool, optional
        Whether to use normalization on outputs from each layer. Default: False.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output, num_nodes), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.
    """
    def __init__(self, 
                num_nodes,
                num_features,
                num_timesteps_input,
                num_timesteps_output,
                nhid = 256,
                dropout = 0.5,
                nhead = 1, 
                num_clusters = [10, 5], 
                use_norm = False):
        
        super(ATMGNN, self).__init__()
        
        self.window = num_timesteps_input
        self.nout = num_timesteps_output
        self.n_nodes = num_nodes
        self.nhid = nhid
        self.nfeat = num_features
        self.nhead = nhead
        self.use_norm = use_norm

        # +--------------------------------------+
        # | Multiresolution Graph Networks (MGN) |
        # +--------------------------------------+

        # Bottom encoder
        self.bottom_encoder = MPNN_Encoder(self.nfeat, nhid, nhid, self.n_nodes, dropout)

        # Number of clusters
        self.num_clusters = num_clusters

        # Multiresolution construction
        self.middle_linear = nn.ModuleList()
        self.middle_encoder = nn.ModuleList()

        for size in self.num_clusters:
            self.middle_linear.append(nn.Linear(nhid, size))
            self.middle_encoder.append(nn.Linear(nhid, nhid))

        # Mixing multiple resolutions together
        self.mix_1 = nn.Linear((len(self.num_clusters) + 1) * nhid, 512)
        self.mix_2 = nn.Linear(512, (len(self.num_clusters) + 1) * nhid)

        # +--------------------------+
        # | Attention/Self-Attention |
        # +--------------------------+

        self.self_attention = nn.MultiheadAttention((len(self.num_clusters) + 1) * nhid, self.nhead, dropout=dropout)
        self.linear_reduction = nn.Linear(self.window, 1)
        
        self.fc1 = nn.Linear((len(self.num_clusters) + 1) * nhid + self.window * self.nfeat, nhid)
        self.fc2 = nn.Linear(nhid, self.nout)
        
        self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU()
        
        
    def forward(self, x, adj, states=None, dynamic_adj=None, **kargs):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input features tensor with shape (batch_size, num_timestamps, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_timesteps_output, num_nodes),
            representing the predicted values for each node over the specified output timesteps.
        """
        lst = list()
        skip = x.view(-1, self.window, self.n_nodes, self.nfeat)
        skip = torch.transpose(skip, 1, 2).reshape(-1, self.window, self.nfeat)

        # +--------------------------------------+
        # | Multiresolution Graph Networks (MGN) |
        # +--------------------------------------+
        x = x.view(-1, self.n_nodes, self.nfeat)

        # All latents
        all_latents = []

        # Bottom encoder
        bottom_latent = self.bottom_encoder(adj, x)

        all_latents.append(bottom_latent)

        # Product of all assignment matrices
        product = None

        # Multiresolution construction
        adj = adj.to_dense()
        latent = bottom_latent
        num_nodes = [self.n_nodes] + self.num_clusters
        for level in range(len(self.num_clusters)):
            size = self.num_clusters[level]

            # Assignment matrix
            assign = self.middle_linear[level](latent.view(-1, bottom_latent.shape[-1]))
            assign = F.gumbel_softmax(assign, tau = 1, hard = True, dim = 1).view(-1, num_nodes[level], assign.shape[-1])

            # Update product
            if level == 0:
                product = assign
            else:
                product = torch.matmul(product, assign)

            # Coarsen node features
            x = torch.matmul(assign.transpose(1, 2), latent)
            x = F.normalize(x, dim = 2)

            # Coarsen the adjacency
            adj = torch.matmul(torch.matmul(assign.transpose(1, 2), adj), assign)
            adj = adj / torch.sum(adj)

            # New latent by graph convolution
            latent = torch.tanh(self.middle_encoder[level](torch.matmul(adj, x))) # h(AX)

            # Extended latent
            extended_latent = torch.matmul(product, latent)
            all_latents.append(extended_latent)

        # Normalization
        if self.use_norm == True:
            for idx in range(len(all_latents)):
                all_latents[idx] = all_latents[idx] / torch.norm(all_latents[idx], p = 2)

        # Concatenate all resolutions
        representation = torch.cat(all_latents, dim = 2)
        x = representation

        # Mixing multiple resolutions
        x = torch.relu(self.mix_1(x.view(-1, x.shape[-1])))
        x = torch.relu(self.mix_2(x))

        # +------------------------------------------------------------------------------------------+
        # | Attention model module, self-attention variant with query, key, and value the same input |
        # +------------------------------------------------------------------------------------------+
        x = x.view(-1, self.window, self.n_nodes, x.shape[-1]) 
        x = torch.transpose(x, 0, 1)
        x = x.contiguous().view(self.window, -1, x.size(3))#self.batch_size*self.n_nodes

        x, _ = self.self_attention(x, x, x)
        x = torch.transpose(x, 0, 2)
        x = self.linear_reduction(x)
        x = x.squeeze()
        x = torch.transpose(x, 0, 1)

        skip = skip.reshape(skip.size(0),-1)
                
        x = torch.cat([x,skip], dim=1)

        x = self.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.relu(self.fc2(x))
        x = x.view(-1, x.shape[-1], self.n_nodes)

        return x
    
    def initialize(self):
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()

    

class MPNN_LSTM(BaseModel):
    """
    Message Passing Neural Network combined with Long Short-Term Memory (MPNN_LSTM)
    
    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features in each timestep of the input data.
    num_timesteps_input : int
        Window size; number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    nhid : int, optional
        Number of hidden units in the LSTM and the number of output channels in GCN layers. Default: 256.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    
    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output, num_nodes), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.

    """
    def __init__(self, 
                num_nodes,
                num_features,
                num_timesteps_input,
                num_timesteps_output,
                nhid = 256,
                dropout = 0.5,
                device = 'cpu'):
        """
        Parameters:
        nfeat (int): Number of features
        nhid (int): Hidden size
        nout (int): Number of output features
        n_nodes (int): Number of nodes
        window (int): Window size
        dropout (float): Dropout rate
        
        Returns:
        x (torch.Tensor): Output of the model
        """
        super(MPNN_LSTM, self).__init__()
        self.window = num_timesteps_input
        self.n_nodes = num_nodes
        self.nhid = nhid
        self.nfeat = num_features
        self.nout = num_timesteps_output
        self.conv1 = GCNConv(self.nfeat, nhid)
        self.conv2 = GCNConv(nhid, nhid)
        
        self.bn1 = nn.BatchNorm2d(nhid) # originally 1d
        self.bn2 = nn.BatchNorm2d(nhid)  
        
        self.rnn1 = nn.LSTM(2*nhid, nhid, 1)
        self.rnn2 = nn.LSTM(nhid, nhid, 1)
        
        self.fc1 = nn.Linear(2*nhid+self.window*self.nfeat, nhid)
        self.fc2 = nn.Linear( nhid, self.nout)
        
        self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU()
        
        
    def forward(self, x, adj, states=None, dynamic_adj=None, **kargs):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input features tensor with shape (batch_size, num_timestamps, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_timesteps_output, num_nodes),
            representing the predicted values for each node over the specified output timesteps.
        """
        lst = list()
        weight = adj.to_sparse().values()
        adj = adj.to_sparse().indices()

        skip = x.view(-1,self.window,self.n_nodes,self.nfeat)
       
        skip = torch.transpose(skip, 1, 2).reshape(-1,self.window,self.nfeat)
        
        x = self.relu(self.conv1(x, adj, edge_weight=weight))
        x = self.bn1(x.view(-1, self.nhid, self.window, self.n_nodes)).view(-1, self.window, self.n_nodes, self.nhid)
        x = self.dropout(x)
        lst.append(x)
        
        x = self.relu(self.conv2(x, adj, edge_weight=weight))
        x = self.bn2(x.view(-1, self.nhid, self.window, self.n_nodes)).view(-1, self.window, self.n_nodes, self.nhid)
        x = self.dropout(x)
        lst.append(x)
        
        x = torch.cat(lst, dim=-1)
        
        # reshape to (seq_len, batch_size , hidden) to fit the lstm
        x = x.view(-1, self.window, self.n_nodes, x.size(3))
        x = torch.transpose(x, 0, 1)
        x = x.contiguous().view(self.window, -1, x.size(3))
   
        x, (hn1, cn1) = self.rnn1(x)
        
        out2, (hn2,  cn2) = self.rnn2(x)
        
        # use the hidden states of both rnns 
        x = torch.cat([hn1[0,:,:],hn2[0,:,:]], dim=1)
        skip = skip.reshape(skip.size(0),-1)
                
        x = torch.cat([x,skip], dim=1)

        x = self.relu(self.fc1(x))
        x = self.dropout(x)
        x = self.relu(self.fc2(x)).squeeze()
        x = x.view(-1, self.nout, self.n_nodes)
        
        return x
    
    def initialize(self):
        self.conv1.reset_parameters()
        self.conv2.reset_parameters()
        self.bn1.reset_parameters()
        self.bn2.reset_parameters()
        self.rnn1.reset_parameters()
        self.rnn2.reset_parameters()
        self.fc1.reset_parameters()
        self.fc2.reset_parameters()



# this code is adopted from https://github.com/v1xerunt/STAN/blob/master/model.py

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn.conv import MessagePassing

from .base import BaseModel


class WeightedSumConv(MessagePassing):
    def __init__(self):
        super(WeightedSumConv, self).__init__(aggr='add')  # 'add' 表示聚合方式为加法
    
    def forward(self, x, edge_index, edge_attr):
        # 触发消息传递
        return self.propagate(edge_index, x=x, edge_attr=edge_attr)
    
    def message(self, x_j, edge_attr):
        return x_j * edge_attr  # 消息 = 源节点特征 * 边权重
    
    def update(self, aggr_out):
        return aggr_out  # 聚合输出


class GATLayer(nn.Module):
    def __init__(self, in_dim, out_dim):
        super(GATLayer, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.fc = nn.Linear(in_dim, out_dim)
        self.attn_fc = nn.Linear(2 * out_dim, 1)
        self.conv = WeightedSumConv()
        self.reset_parameters()

    def reset_parameters(self):
        gain = nn.init.calculate_gain('relu')
        nn.init.xavier_normal_(self.fc.weight, gain=gain)
        nn.init.xavier_normal_(self.attn_fc.weight, gain=gain)

    def forward(self, adj, h):
        z = self.fc(h)

        sparse_adj = adj.to_sparse_coo()
        src_index = sparse_adj.indices()[0]
        dst_index = sparse_adj.indices()[1]
        try:
            att_feat = torch.cat([z[:,src_index],z[:,dst_index]], dim = -1)
        except:
            att_feat = torch.cat([z[src_index],z[dst_index]], dim = -1)
        att_edge = F.leaky_relu(self.attn_fc(att_feat))

        output = self.conv(z, sparse_adj.indices(), att_edge)

        return output  #self.g.ndata.pop('h')

class MultiHeadGATLayer(nn.Module):
    def __init__(self, in_dim, out_dim, num_heads, merge='cat'):
        super(MultiHeadGATLayer, self).__init__()
        self.heads = nn.ModuleList()
        for i in range(num_heads):
            self.heads.append(GATLayer(in_dim, out_dim))
        self.merge = merge

    def forward(self, adj, h):
        head_outs = [attn_head(adj, h) for attn_head in self.heads]
        if self.merge == 'cat':
            return torch.cat(head_outs, dim=-1)
        else:
            return torch.mean(torch.stack(head_outs))
        
class STAN(BaseModel):
    """
    Spatio-Temporal Attention Network (STAN)

    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features per node per timestep.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    population : float, optional
        Total population considered in the model. Default: 1e10.
    gat_dim1 : int
        Dimension of the output space for the first GAT layer. Default: 32.
    gat_dim2 : int
        Dimension of the output space for the second GAT layer. Default: 32.
    gru_dim : int
        Dimension of the hidden state for the GRU layer. Default: 32.
    num_heads : int
        Number of attention heads in the first GAT layer. Default: 1.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    Returns
    -------
    tuple of torch.Tensor
        A tuple containing two tensors, each of shape (batch_size, num_timesteps_output, num_nodes, 1), representing the predicted values for newly infected and recovered individuals respectively for each node over future timesteps. The second tensor contains the physical model predictions.
    """
    def __init__(self, 
                 num_nodes, 
                 num_features, 
                 num_timesteps_input, 
                 num_timesteps_output, 
                 population=1e10, 
                 gat_dim1=32, 
                 gat_dim2=32, 
                 gru_dim=32, 
                 num_heads=1, 
                 device = 'cpu'):
        super(STAN, self).__init__()
        self.n_nodes = num_nodes
        self.nfeat = num_features
        self.history = num_timesteps_input
        self.horizon = num_timesteps_output
        self.pop = population
        
        self.layer1 = MultiHeadGATLayer(self.nfeat*self.history, gat_dim1, num_heads)
        self.layer2 = MultiHeadGATLayer(gat_dim1 * num_heads, gat_dim2, 1)

        self.pred_window = self.horizon
        # self.gru = nn.GRUCell(gat_dim2, gru_dim)
        self.gru = nn.GRU(gat_dim2, gru_dim, num_layers=1, dropout=0.5)
    
        self.nn_res_I = nn.Linear(gru_dim+2, self.horizon)
        self.nn_res_R = nn.Linear(gru_dim+2, self.horizon)

        self.nn_res_sir = nn.Linear(gru_dim+2, 2)
        
        self.hidden_dim2 = gat_dim2
        self.gru_dim = gru_dim
        self.device = device

    def forward(self, X, adj, states, dynamic_adj=None, N = None, h = None):
        """
        Parameters
        ----------
        X : torch.Tensor
            Input feature tensor with shape (batch_size, num_timesteps_input, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix with shape (num_nodes, num_nodes).
        states : torch.Tensor
            States of the nodes, with the same shape as X, containing current infection and recovery data.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix, with shape similar to adj but possibly varying over time. Default: None.
        N : float, optional
            Total population considered in the model. Default: None.
        h : torch.Tensor, optional
            Hidden states for the GRU layer, used if provided. Default: None.

        Returns
        -------
        tuple of torch.Tensor
            A tuple containing two tensors:
            1. Predicted new infections and recoveries, shape (batch_size, num_timesteps_output, num_nodes, 2).
            2. Physical model predictions based on current states, shape (batch_size, num_timesteps_output, num_nodes, 2).
        """
        last_diff_I = X[:, -1, :, 1].unsqueeze(2)
        last_diff_R = X[:, -1, :, 2].unsqueeze(2)
        X = X.transpose(1,2).flatten(2,3)
        

        # if h is None:
        #     h = torch.zeros(X.shape[0], self.gru_dim).to(self.device)
        #     gain = nn.init.calculate_gain('relu')
        #     nn.init.xavier_normal_(h, gain=gain)  

        new_I = []
        new_R = []
        phy_I = []
        phy_R = []
        self.alpha_list = []
        self.beta_list = []
        self.alpha_scaled = []
        self.beta_scaled = [] 

        cur_h = self.layer1(adj, X)
        cur_h = F.elu(cur_h)
        cur_h = self.layer2(adj, cur_h)
        cur_h = F.elu(cur_h)
        
        h, last_h = self.gru(cur_h)
        hc = torch.cat((h, last_diff_I, last_diff_R) , dim = -1)

        pred_I = self.nn_res_I(hc).transpose(1,2).unsqueeze(-1)
        pred_R = self.nn_res_R(hc).transpose(1,2).unsqueeze(-1)

        pred_res = self.nn_res_sir(hc)

        alpha = torch.sigmoid(pred_res[..., 0])
        beta =  torch.sigmoid(pred_res[..., 1])

        phy_I = []
        phy_R = []
        if N is None:
            N = self.pop
        for i in range(self.horizon):
            last_I = states[:,-1,:, 1] if i == 0 else last_I + dI.detach()
            last_R = states[:,-1,:, 2] if i == 0 else last_R + dR.detach()

            last_S = N - last_I - last_R
            
            dI = alpha * last_I * (last_S/N) - beta * last_I
            dR = beta * last_I
            phy_I.append(dI)
            phy_R.append(dR)
        phy_I = torch.stack(phy_I).to(self.device).transpose(1,0).unsqueeze(-1)
        phy_R = torch.stack(phy_R).to(self.device).transpose(1,0).unsqueeze(-1)

        return torch.cat([pred_I, pred_R], dim=-1), torch.cat([phy_I, phy_R], dim=-1)
    
    def initialize(self):
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()



# this code is adopted from https://github.com/Xiefeng69/EpiGNN/blob/main/src/models.py

import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import Parameter

from .base import BaseModel

def getLaplaceMat(batch_size, m, adj):
    i_mat = torch.eye(m).to(adj.device)
    i_mat = i_mat.unsqueeze(0)
    o_mat = torch.ones(m).to(adj.device)
    o_mat = o_mat.unsqueeze(0)
    i_mat = i_mat.expand(batch_size, m, m)
    o_mat = o_mat.expand(batch_size, m, m)
    adj = torch.where(adj>0, o_mat, adj)
    '''
    d_mat = torch.bmm(adj, adj.permute(0, 2, 1))
    d_mat = torch.where(i_mat>0, d_mat, i_mat)
    print('d_mat version 1', d_mat)
    '''
    d_mat_in = torch.sum(adj, dim=1)
    d_mat_out = torch.sum(adj, dim=2)
    d_mat = torch.sum(adj, dim=2) # attention: dim=2
    d_mat = d_mat.unsqueeze(2)
    d_mat = d_mat + 1e-12
    #d_mat = torch.pow(d_mat, -0.5) if is 1/2
    d_mat = torch.pow(d_mat, -1)
    d_mat = d_mat.expand(d_mat.shape[0], d_mat.shape[1], d_mat.shape[1])
    d_mat = i_mat * d_mat

    # laplace_mat = d_mat * adj * d_mat
    laplace_mat = torch.bmm(d_mat, adj)
    #laplace_mat = torch.bmm(laplace_mat, d_mat)
    return laplace_mat


class GraphConvLayer(nn.Module):
    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.Tensor(in_features, out_features))
        self.act = nn.ELU()
        nn.init.xavier_uniform_(self.weight)

        if bias:
            self.bias = Parameter(torch.Tensor(out_features))
            stdv = 1. / math.sqrt(self.bias.size(0))
            self.bias.data.uniform_(-stdv, stdv)
        else:
            self.register_parameter('bias', None)

    def forward(self, feature, adj):
        support = torch.matmul(feature, self.weight)
        output = torch.matmul(adj, support)

        if self.bias is not None:
            return self.act(output + self.bias)
        else:
            return self.act(output)

class GraphLearner(nn.Module):
    def __init__(self, hidden_dim, tanhalpha=1):
        super().__init__()
        self.hid = hidden_dim
        self.linear1 = nn.Linear(self.hid, self.hid)
        self.linear2 = nn.Linear(self.hid, self.hid)
        self.alpha = tanhalpha

    def forward(self, embedding):
        # embedding [batchsize, hidden_dim]
        nodevec1 = self.linear1(embedding)
        nodevec2 = self.linear2(embedding)
        nodevec1 = self.alpha * nodevec1
        nodevec2 = self.alpha * nodevec2
        nodevec1 = torch.tanh(nodevec1)
        nodevec2 = torch.tanh(nodevec2)
        
        adj = torch.bmm(nodevec1, nodevec2.permute(0, 2, 1))-torch.bmm(nodevec2, nodevec1.permute(0, 2, 1))
        adj = self.alpha * adj
        adj = torch.relu(torch.tanh(adj))
        return adj

class ConvBranch(nn.Module):
    def __init__(self,
                 m,
                 in_channels,
                 out_channels,
                 kernel_size,
                 dilation_factor,
                 hidP=1,
                 isPool=True):
        super().__init__()
        self.m = m
        self.isPool = isPool
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=(kernel_size,1), dilation=(dilation_factor,1))
        self.batchnorm = nn.BatchNorm2d(out_channels)
        if self.isPool:
            self.pooling = nn.AdaptiveMaxPool2d((hidP, m))
        #self.activate = nn.Tanh()
    
    def forward(self, x):
        batch_size = x.shape[0]
        x = self.conv(x)
        x = self.batchnorm(x)
        if self.isPool:
            x = self.pooling(x)
        x = x.view(batch_size, -1, self.m)
        return x

class RegionAwareConv(nn.Module):
    def __init__(self, nfeat, P, m, k, hidP, dilation_factor=2):
        super(RegionAwareConv, self).__init__()
        self.nfeat = nfeat
        self.P = P
        self.m = m
        self.k = k
        self.hidP = hidP
        self.conv_l1 = ConvBranch(m=self.m, in_channels=nfeat, out_channels=self.k, kernel_size=3, dilation_factor=1, hidP=self.hidP)
        self.conv_l2 = ConvBranch(m=self.m, in_channels=nfeat, out_channels=self.k, kernel_size=5, dilation_factor=1, hidP=self.hidP)
        self.conv_p1 = ConvBranch(m=self.m, in_channels=nfeat, out_channels=self.k, kernel_size=3, dilation_factor=dilation_factor, hidP=self.hidP)
        self.conv_p2 = ConvBranch(m=self.m, in_channels=nfeat, out_channels=self.k, kernel_size=5, dilation_factor=dilation_factor, hidP=self.hidP)
        self.conv_g = ConvBranch(m=self.m, in_channels=nfeat, out_channels=self.k, kernel_size=self.P, dilation_factor=1, hidP=None, isPool=False)
        self.activate = nn.Tanh()
    
    def forward(self, x):
        x = x.view(-1, self.nfeat, self.P, self.m)
        batch_size = x.shape[0]
        # local pattern
        x_l1 = self.conv_l1(x)
        x_l2 = self.conv_l2(x)
        x_local = torch.cat([x_l1, x_l2], dim=1)
        # periodic pattern
        x_p1 = self.conv_p1(x)
        x_p2 = self.conv_p2(x)
        x_period = torch.cat([x_p1, x_p2], dim=1)
        # global
        x_global = self.conv_g(x)
        # concat and activate
        x = torch.cat([x_local, x_period, x_global], dim=1).permute(0, 2, 1)
        x = self.activate(x)
        return x


class EpiGNN(BaseModel):
    """
    Epidemiological Graph Neural Network (EpiGNN)

    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features per node per timestep.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    k : int, optional
        Number of local neighborhoods to consider in the graph learning layer. Default: 8.
    hidA : int, optional
        Dimension of attention in the model. Default: 64.
    hidR : int, optional
        Dimension of hidden layers in the recurrent neural network part. Default: 40.
    hidP : int, optional
        Dimension of positional encoding in the model. Default: 1.
    n_layer : int, optional
        Number of layers in the graph neural network. Default: 2.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output, num_nodes), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.
    """
    def __init__(self, 
                num_nodes, 
                num_features, 
                num_timesteps_input,
                num_timesteps_output, 
                k = 8, 
                hidA = 64, 
                hidR = 40, 
                hidP=1, 
                n_layer = 2, 
                dropout = 0.5, 
                device = 'cpu'):
        super().__init__()
        # arguments setting
        self.device = device
        self.nfeat = num_features
        self.m = num_nodes
        self.w = num_timesteps_input
        self.droprate = dropout
        self.hidR = hidR
        self.hidA = hidA
        self.hidP = hidP
        self.k =k
        self.n = n_layer
        self.dropout = nn.Dropout(self.droprate)

        # unavailable settings
        self.res = 0
        self.hw = 0
        if self.hw > 0:
            self.highway = nn.Linear(self.hw, 1)
        self.extra = False

        # if args.extra:
        #     self.extra = True
        #     self.external = data.external
        # else:
        #     self.extra = False

        # Feature embedding
        self.hidR = self.k*4*self.hidP + self.k
        self.backbone = RegionAwareConv(nfeat = num_features, P=self.w, m=self.m, k=self.k, hidP=self.hidP)

        # global
        self.WQ = nn.Linear(self.hidR, self.hidA)
        self.WK = nn.Linear(self.hidR, self.hidA)
        self.leakyrelu = nn.LeakyReLU(inplace=True)
        self.t_enc = nn.Linear(1, self.hidR)

        # local
        self.s_enc = nn.Linear(1, self.hidR)

        # external resources
        self.external_parameter = nn.Parameter(torch.FloatTensor(self.m, self.m), requires_grad=True)

        # Graph Generator and GCN
        self.d_gate = nn.Parameter(torch.FloatTensor(self.m, self.m), requires_grad=True)
        self.graphGen = GraphLearner(self.hidR)
        self.GNNBlocks = nn.ModuleList([GraphConvLayer(in_features=self.hidR, out_features=self.hidR) for i in range(self.n)])
        #self.GCNBlock1 = GraphConvLayer(in_features=self.hidR, out_features=self.hidR)
        #self.GCNBlock2 = GraphConvLayer(in_features=self.hidR, out_features=self.hidR)

        # prediction
        if self.res == 0:
            self.output = nn.Linear(self.hidR*2, num_timesteps_output)
        else:
            self.output = nn.Linear(self.hidR*(self.n+1), num_timesteps_output)

        self.init_weights()
     
    def init_weights(self):
        for p in self.parameters():
            if p.data.ndimension() >= 2:
                nn.init.xavier_uniform_(p.data) # best
            else:
                stdv = 1. / math.sqrt(p.size(0))
                p.data.uniform_(-stdv, stdv)
    
    def forward(self, X, adj, states=None, dynamic_adj=None, index=None):
        """
        Parameters
        ----------
        X : torch.Tensor
            Input features tensor with shape (batch_size, num_timesteps_input, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_timesteps_output, num_nodes),
            representing the predicted values for each node over the specified output timesteps.
        """
        adj = adj.bool().float()
        #print(index.shape) batch_size
        batch_size = X.shape[0] # batchsize, w, m, feat

        # step 1: Use multi-scale convolution to extract feature embedding (SEFNet => RAConv).
        temp_emb = self.backbone(X)

        # step 2: generate global transmission risk encoding.
        query = self.WQ(temp_emb) # batch, N, hidden
        query = self.dropout(query)
        key = self.WK(temp_emb)
        key = self.dropout(key)
        attn = torch.bmm(query, key.transpose(1, 2))
        #attn = self.leakyrelu(attn)
        attn = F.normalize(attn, dim=-1, p=2, eps=1e-12)
        attn = torch.sum(attn, dim=-1)
        attn = attn.unsqueeze(2)
        t_enc = self.t_enc(attn)
        t_enc = self.dropout(t_enc)

        # step 3: generate local transmission risk encoding.
        # print(self.degree.shape) [self.m]
        d = adj.sum(0).unsqueeze(1)
        s_enc = self.s_enc(d)
        s_enc = self.dropout(s_enc)

        # Three embedding fusion.
        feat_emb = temp_emb + t_enc + s_enc

        # step 4: Region-Aware Graph Learner
        # load external resource
        if self.extra:
            extra_adj_list=[]
            zeros_mt = torch.zeros((self.m, self.m)).to(self.adj.device)
            #print(self.external.shape)
            for i in range(batch_size):
                offset = 20
                if i-offset>=0:
                    idx = i-offset
                    extra_adj_list.append(self.external[index[i],:,:].unsqueeze(0))
                else:
                    extra_adj_list.append(zeros_mt.unsqueeze(0))
            extra_info = torch.concat(extra_adj_list, dim=0) # [1872, 52]
            extra_info = extra_info
            #print(extra_info.shape) # batch_size, self.m self.m
            external_info = torch.mul(self.external_parameter, extra_info)
            external_info = F.relu(external_info)
            #print(self.external_parameter)

        # apply Graph Learner to generate a graph
        d_mat = torch.mm(d, d.permute(1, 0))
        d_mat = torch.mul(self.d_gate, d_mat)
        d_mat = torch.sigmoid(d_mat)
        spatial_adj = torch.mul(d_mat, adj)
        adj = self.graphGen(temp_emb)
        
        # if additional information => fusion
        if self.extra:
            adj = adj + spatial_adj + external_info
        else:
            adj = adj + spatial_adj

        # get laplace adjacent matrix
        laplace_adj = getLaplaceMat(batch_size, self.m, adj)
        
        # Graph Convolution Network
        node_state = feat_emb
        node_state_list = []
        for layer in self.GNNBlocks:
            node_state = layer(node_state, laplace_adj)
            node_state = self.dropout(node_state)
            node_state_list.append(node_state)
        '''
        if self.res == 1:
            node_state = torch.cat(node_state_list, dim=-1)
        
        node_state = self.GCNBlock1(feat_emb, laplace_adj)
        node_state = self.dropout(node_state)
        node_state = self.GCNBlock2(node_state, laplace_adj)
        node_state = self.dropout(node_state)
        '''

        # Final prediction
        node_state = torch.cat([node_state, feat_emb], dim=-1)
        res = self.output(node_state)
        # highway means autoregressive model
        if self.hw > 0:
            z = X[:, -self.hw:, :]
            z = z.permute(0, 2, 1).contiguous().view(-1, self.hw)
            z = self.highway(z)
            z = z.view(-1, self.m)
            res = res + z
        
        return res.transpose(1,2)
    
    def initialize(self):
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()




class Dataset():
    def __init__(   self,

                 ):
        pass

    def download(self):
        """Download selected files of the dataset."""

    def save(self):
        """Save current dataset."""

    def get_slice(self, timestamp):
        """Get a slice of graph or time series"""
    




import torch
import numpy as np
import random
import math
import time

from .utils import edge_to_adj

def get_random_graph(num_nodes=None, connect_prob=None, block_sizes=None, num_edges=None, graph_type='erdos_renyi'):
    """
    Generates a random static graph using one of the supported graph types: Erdos-Renyi, Stochastic Blockmodel, or Barabasi-Albert.

    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    connect_prob : float, optional
        Probability of edge creation (for Erdos-Renyi and Stochastic Blockmodel graphs).
    block_sizes : list of int, optional
        Sizes of blocks (for Stochastic Blockmodel graph).
    num_edges : int, optional
        Number of edges (for Barabasi-Albert graph).
    graph_type : str
        Type of graph to generate. Options are `'erdos_renyi'`, `'stochastic_blockmodel'`, `'barabasi_albert'`. Default is `'erdos_renyi'`.

    Returns
    -------
    torch.Tensor
        Adjacency matrix of the generated graph.
    """
    import torch_geometric
    if graph_type == 'erdos_renyi':
        edges = torch_geometric.utils.erdos_renyi_graph(num_nodes, connect_prob)
    elif graph_type == 'stochastic_blockmodel':
        edges = torch_geometric.utils.stochastic_blockmodel_graph(block_sizes, connect_prob)
    elif graph_type == 'barabasi_albert':
        edges = torch_geometric.utils.barabasi_albert_graph(num_nodes, num_edges)
    else:
        raise NameError("grap type not supported")
    
    adj = edge_to_adj(edges, num_nodes)
    return adj

def get_graph_from_features(features, adj=None, G=1):
    """
    Generate a graph from node features using cosine similarity.

    This function generates a graph where each edge weight is computed based on the cosine similarity
    between the feature vectors of the connected nodes. If an adjacency matrix is provided, the cosine 
    similarity is adjusted by the corresponding entry in the adjacency matrix.

    Parameters
    ----------
    features : torch.Tensor
        A tensor of shape (num_nodes, feat_dim) where num_nodes is the number of nodes and feat_dim is 
        the dimensionality of the feature vectors.
    adj : torch.Tensor, optional
        A tensor of shape (num_nodes, num_nodes) representing the adjacency matrix, where adj[i, j] 
        denotes the distance or weight between node i and node j. If None, the cosine similarity is 
        used directly as the edge weight. Default is None.

    Returns
    -------
    torch.Tensor
        A tensor of shape (num_nodes, num_nodes) representing the generated graph's adjacency matrix, 
        where each entry [i, j] contains the adjusted cosine similarity between nodes i and j.
    """
    n_nodes = len(features)
    graph = torch.zeros(n_nodes, n_nodes)
    for i in range(n_nodes):
        for j in range(n_nodes):
            if adj is not None:
                graph[i, j] = (torch.nn.functional.cosine_similarity(features[i], features[j], dim=0)/adj[i,j]).item()
            else:
                graph[i, j] = torch.nn.functional.cosine_similarity(features[i], features[j], dim=0).item()
    return graph

class Gravity_model(object):
    """
    A class to represent a gravity model for mobility analysis in an epidemic context.

    The gravity model is used to estimate the influence and mobility between nodes (e.g., cities or regions)
    based on their populations and the distance between them, according to the formula:

    .. math::

        e_{v,w} = \\frac{N_v^{\\rho} N_w^{\\theta}}{\\exp(d_{vw} / \\delta)},\\forall v, w \\in \\mathcal{V}
        
        
    """
    def __init__(self, source, target, s):
        """
        Initialize the Gravity_model with specified parameters.

        Parameters
        ----------
        source : float
            Exponent for the source node population :math:`\\rho` in the gravity model.
        target : float
            Exponent for the target node population :math:`\\theta` in the gravity model.
        s : float
            Scaling factor for the distance :math:`\\delta` in the gravity model.
        """
        self.source = source
        self.target = target
        self.s = s
    
    def get_influence(self, source_population, target_population, distance):
        """
        Calculate the influence between a source and a target node based on their populations and the distance between them.

        Parameters
        ----------
        source_population : float
            Population of the source node :math:`N_v`.
        target_population : float
            Population of the target node :math:`N_w`.
        distance : float
            Distance between the source and target nodes :math:`d_{vw}`.

        Returns
        -------
        float
            The calculated influence between the source and target nodes.
        """
        return (source_population**self.source * target_population**self.target) / torch.exp(distance / self.s)

    def get_mobility_graph(self, node_populations, distance_graph):
        """
        Generate a mobility graph based on node populations and a distance graph.

        Parameters
        ----------
        node_populations : torch.Tensor
            A tensor of shape (num_nodes,) representing the population of each node.
        distance_graph : torch.Tensor
            A tensor of shape (num_nodes, num_nodes) representing the distances between nodes.

        Returns
        -------
        torch.Tensor
            A tensor of shape (num_nodes, num_nodes) representing the mobility graph, where each entry [i, j]
            represents the mobility influence from node i to node j.
        """
        num_nodes = node_populations.shape[0]
        mobility_graph = torch.zeros(num_nodes, num_nodes)

        for source in range(num_nodes):
            for target in range(num_nodes):
                mobility_graph[source, target] = (node_populations[source]**self.source * node_populations[target]**self.target) / torch.exp(distance_graph[source, target] / self.s)
        return mobility_graph
        

class Time_geo(object):
    """
    Models spatial-temporal movement patterns of individuals across different regions, simulating movement based on
    temporal rhythms, personal habits, and regional exploration probabilities. It includes mechanisms to simulate
    individual trajectories, predict location changes, and calculate movement probabilities based on historical data.

    Parameters
    ----------
    region_input : ndarray
        Input array of region coordinates.
    pop_input : ndarray
        Population distribution across regions.
    p_t_raw : ndarray, optional
        Raw probability distribution of time slots for movement. Default: Loads from "epilearn/data/rhythm.npy".
    pop_num : int, optional
        Number of individuals in the population to simulate. Default: 7.
    time_slot : int, optional
        Time resolution in minutes for one slot. Default: 10.
    rho : float, optional
        Controls the exploration probability for other regions. Default: 0.6.
    gamma : float, optional
        Attenuation parameter for exploration probability. Default: 0.41.
    alpha : float, optional
        Controls the exploration depth. Default: 1.86.
    n_w : float, optional
        Average number of tours based on home per week. Default: 6.1.
    beta1 : float, optional
        Dwell rate. Default: 3.67.
    beta2 : float, optional
        Burst rate, affecting movement speed. Default: 10.
    simu_slot : int, optional
        Total number of simulation slots. Default: 144.

    Returns
    -------
    list of dicts
        Each dictionary contains the movement trace and other metrics for an individual simulated over the simu_slot duration.
    """
    def __init__(self, region_input, pop_input, p_t_raw=None, pop_num=7, time_slot=10, rho=0.6, gamma=0.41, alpha=1.86, n_w=6.1, beta1=3.67, beta2=10, simu_slot=144):
        
        super().__init__()
        self.time_slot = time_slot # time resolution is half an hour
        self.rho = rho # it controls the exploration probability for other regions
        self.gamma = gamma # it is the attenuation parameter for exploration probability
        self.alpha = alpha # it controls the exploration depth
        self.n_w = n_w # it is the average number of tour based on home a week.
        self.beta1 = beta1 # dwell rate
        self.beta2 = beta2 # burst rate
        self.simu_slot = simu_slot
        self.pop_num = pop_num

        self.sample_region = region_input
        p_t_raw = p_t_raw if p_t_raw is not None else np.load("epilearn/data/rhythm.npy", allow_pickle=True)        
        self.p_t = np.array(p_t_raw).reshape(-1, (time_slot // 10)).sum(axis=1)
        self.region_num = self.sample_region.shape[0]
        self.home_location = np.random.choice(len(pop_input), pop_num, p=pop_input)
        self.pop_info = self.trace_simulate()

    def distance(self, p1, p2):#caculate distance
        return math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)
    
    def get_p_t(self, now_time):
        now_time_tup = time.localtime(float(now_time))
        i = int((now_time_tup.tm_wday * 24 * 60 + now_time_tup.tm_hour * 60 + now_time_tup.tm_min) / self.time_slot)
        return self.p_t[i]
    
    def predict_next_place_time(self, p_t_value, current_location_type):
        p1 = 1 - self.n_w * p_t_value
        p2 = 1 - self.beta1 * self.n_w * p_t_value
        p3 = self.beta2 * self.n_w * p_t_value
        location_is_change = 0
        new_location_type = 'undefined'
        if current_location_type == 'home':
            if random.uniform(0, 1) <= p1:
                new_location_type = 'home'
                location_is_change = 0
            else:
                new_location_type = 'other'
                location_is_change = 1
        elif current_location_type == 'other':
            p = random.uniform(0, 1)
            if p <= p2:
                new_location_type = 'other'
                location_is_change = 0
            elif random.uniform(0, 1) <= p3:
                new_location_type = 'other'
                location_is_change = 1
            else:
                new_location_type = 'home'
                location_is_change = 1
        if new_location_type == 'home':
            return 0, location_is_change
        else:
            return 2, location_is_change
        
    def negative_pow(self, k):
        p_k = {}
        for i, region in enumerate(k, 1):
            p_k[region[0]] = i ** (-self.alpha)
        temp = sum(p_k.values())
        for key in p_k:
            p_k[key] = p_k[key] / temp
        return p_k

    def predict_next_place_location_simplify(self, P_new, region_history, current_region, home_region):
        rp = random.uniform(0, 1)
        prob_accum, next_region = 0, 0
        if random.uniform(0, 1) < P_new:
            # explore; the explore distance is depend on history->delta_r
            length = {}
            for i, cen in enumerate(self.sample_region):
                if i in region_history:
                    continue
                length[i] = self.distance(cen, self.sample_region[current_region])
            try:
                del length[home_region]
                del length[current_region]
            except KeyError:
                pass
            k = sorted(length.items(), key=lambda x: x[1], reverse=False)
            p_k = self.negative_pow(k)
            for i, key in enumerate(p_k):
                prob_accum += p_k[key]
                if prob_accum > rp:
                    next_region = key
                    region_history[key] = 1
                    break
                else:
                    continue
        else:
            # return
            region_history_sum = sum(region_history.values())
            for key in region_history:
                prob_accum += region_history[key]/region_history_sum
                if rp < prob_accum:
                    next_region = key
                    region_history[key] += 1
                    break
        return next_region

    def predict_next_place_location(self, region_history, current_location, home_region):
        s = len(region_history.values())
        p_new = self.rho * s ** (-self.gamma) if s != 0 else 1
        return self.predict_next_place_location_simplify(p_new, region_history, current_location, home_region)

    def individual_trace_simulate(self, info, start_time, simu_slot):
        current_location_type = 'home'
        simu_trace = [[info['home'],start_time]]
        for i in range(simu_slot - 1):
            # pt is the currently move based probability
            now_time = (i+1) * 60 * self.time_slot + start_time
            p_t_value = self.get_p_t(now_time)
            now_type, location_change = self.predict_next_place_time(p_t_value, current_location_type)
            if location_change == 1:
                current_location = simu_trace[-1][0]
                if now_type == 0:
                    next_location = info['home']
                    current_location_type = 'home'
                else:
                    next_location = self.predict_next_place_location(info['region_history'], current_location, info['home'])
                    current_location_type = 'other'
                info['feature']['move_num'] += 1
                info['feature']['move_distance'] += self.distance(self.sample_region[next_location], self.sample_region[current_location])
            else:
                next_location = simu_trace[-1][0]
            simu_trace.append([next_location, now_time])
        return simu_trace
    
    def trace_simulate(self):
        """
        Generates simulated movement traces for each individual in the population. The simulation captures
        the dynamics of movement based on initial conditions and predefined movement parameters over a set
        number of time slots.

        Returns
        -------
        list of dicts
            Each dictionary contains detailed information and metrics for an individual's simulated trace, including
            home location, total number of movements, cumulative distance traveled, and the sequence of visited
            locations across the simulation period.
        """
        pop_info = []
        for i in range(self.pop_num):
            pop_info.append({'n_w': self.n_w, 'beta1': self.beta1, 'beta2': self.beta2, 'home': self.home_location[i],
                            'feature': {'move_num': 0, 'move_distance': 0}, 'region_history': {}})
            pop_info[i]['trace'] = np.array(self.individual_trace_simulate(pop_info[i], 1621785600, self.simu_slot))
        return pop_info



import torch
from torch import nn


class SIR(nn.Module):
    """
    Susceptible-Infected-Recovered (SIR) Model

    Parameters
    ----------
    horizon : int, optional
        Number of future time steps to simulate. If None, a single step is simulated unless overridden in the forward method.
    infection_rate : float, optional
        Initial infection rate parameter, representing the rate at which susceptible individuals become infected. Default: 0.01.
    recovery_rate : float, optional
        Initial recovery rate parameter, representing the rate at which infected individuals recover. Default: 0.038.
    population : int, optional
        Total population considered in the model. If None, the sum of the initial conditions (susceptible, infected, recovered) is used as the total population.

    Attributes
    ----------
    beta : torch.nn.Linear
        Linear layer with no bias to model the infection rate dynamically.
    gamma : torch.nn.Linear
        Linear layer with no bias to model the recovery rate dynamically.

    Returns
    -------
    torch.Tensor
        A tensor of shape (horizon, 3), representing the predicted number of susceptible, infected, and recovered individuals at each timestep.
        Each row corresponds to a timestep, with the columns representing susceptible, infected, and recovered counts respectively.

    """
    def __init__(self, horizon=None, infection_rate=0.01, recovery_rate=0.038, population=None):
        super(SIR, self).__init__()
        self.pop = population
        self.horizon = horizon

        self.beta = nn.Linear(1, 1, bias = False)
        self.gamma = nn.Linear(1, 1, bias = False)

        if infection_rate is not None: 
            self.beta.weight.data = torch.FloatTensor([infection_rate])

        if recovery_rate is not None:
            self.gamma.weight.data = torch.FloatTensor([recovery_rate])
        
    def forward(self, x, steps=1):
        """
        Parameters
        ----------
        x : torch.Tensor
            The initial condition tensor for the model. Expected shape is (3,), where the elements represent the 
            number of susceptible (S), infected (I), and recovered (R) individuals respectively.
        steps : int, optional
            Number of future time steps to simulate. If `horizon` is specified during initialization and not None,
            it overrides this parameter. Default is 1 if `horizon` is None.

        Returns
        -------
        torch.Tensor
            A tensor of shape (steps, 3), representing the predicted number of susceptible, infected, and recovered 
            individuals at each timestep. Each row corresponds to a timestep, with the columns representing 
            susceptible, infected, and recovered counts respectively.
        """
        if self.pop is not None:
            pop = self.pop
        else:
            pop = x.sum()
        if self.horizon is not None:
            steps = self.horizon

        assert len(x.shape) == 1 and x.shape[0] == 3

        output = torch.zeros(steps* 3, dtype=torch.float, requires_grad = False).reshape(steps, 3)
        output.data[0] = x.data

        for i in range(1, steps):
            output.data[i][0] = output.data[i-1][0] - self.beta((output.data[i-1][0] * output.data[i-1][1]).unsqueeze(0))/pop
            output.data[i][1] = output.data[i-1][1] + self.beta((output.data[i-1][0] * output.data[i-1][1]).unsqueeze(0))/pop - self.gamma(output.data[i-1][1].unsqueeze(0)) 
            output.data[i][2] = output.data[i-1][2] + self.gamma(output.data[i-1][1].unsqueeze(0))
        return output



class SIS(nn.Module):
    """
    Susceptible-Infected-Susceptible (SIS) Model

    Parameters
    ----------
    horizon : int, optional
        Number of future time steps to simulate. If None, a single step is simulated unless overridden in the forward method.
    infection_rate : float, optional
        Infection rate parameter, representing the rate at which susceptible individuals become infected. If None, must be initialized separately.
    recovery_rate : float, optional
        Recovery rate parameter, representing the rate at which infected individuals recover and return to the susceptible state. If None, must be initialized separately.
    population : int, optional
        Total population considered in the model. If None, the sum of the initial conditions (susceptible and infected) is used as the total population.

    Attributes
    ----------
    beta : torch.nn.Linear
        Linear layer with no bias to model the infection rate dynamically.
    gamma : torch.nn.Linear
        Linear layer with no bias to model the recovery rate dynamically.

    Returns
    -------
    torch.Tensor
        A tensor of shape (horizon, 2), representing the predicted number of susceptible and infected individuals at each timestep.
        Each row corresponds to a timestep, with the columns representing the susceptible and infected counts respectively.

    """
    def __init__(self, horizon=None, infection_rate = None, recovery_rate = None, population = None):
        super(SIS, self).__init__()
        self.pop = population
        self.horizon = horizon

        self.beta = nn.Linear(1, 1, bias = False)
        self.gamma = nn.Linear(1, 1, bias = False)

        if infection_rate is not None:
            self.beta.weight.data = torch.FloatTensor([infection_rate])

        if recovery_rate is not None:
            self.gamma.weight.data = torch.FloatTensor([recovery_rate])
        
    def forward(self, x, steps=1):
        """
        Parameters
        ----------
        x : torch.Tensor
            The initial condition tensor for the model, representing the initial numbers of susceptible (S) and infected (I) individuals.
            Expected shape is (2,), where x[0] is the number of susceptible and x[1] is the number of infected individuals at the start.
        steps : int, optional
            Number of future time steps to simulate. If `horizon` is specified during initialization and not None, it overrides this parameter.
            Default is 1 if `horizon` is None.

        Returns
        -------
        torch.Tensor
            A tensor of shape (steps, 2), representing the predicted number of susceptible and infected individuals at each timestep.
            Each row corresponds to a timestep, with the first column representing susceptible and the second column representing infected counts.
        """
        if self.pop is not None:
            pop = self.pop
        else:
            pop = x.sum()
        if self.horizon is not None:
            steps = self.horizon

        assert len(x.shape) == 1 and x.shape[0] == 2

        output = torch.zeros(steps* 2, dtype=torch.float, requires_grad = False).reshape(steps, 2)
        output.data[0] = x.data

        for i in range(1, steps):
            output.data[i][0] = output.data[i-1][0] - self.beta((output.data[i-1][0] * output.data[i-1][1]).unsqueeze(0))/pop + self.gamma(output.data[i-1][1].unsqueeze(0))
            output.data[i][1] = output.data[i-1][1] + self.beta((output.data[i-1][0] * output.data[i-1][1]).unsqueeze(0))/pop - self.gamma(output.data[i-1][1].unsqueeze(0)) 
            

        return output

class SEIR(nn.Module):
    """
    Susceptible-Exposed-Infected-Recovered (SEIR) Model

    Parameters
    ----------
    horizon : int, optional
        Number of future time steps to simulate. If None, a single step is simulated unless overridden in the forward method.
    infection_rate : float, optional
        Infection rate parameter, representing the rate at which susceptible individuals become exposed. If None, must be initialized separately.
    recovery_rate : float, optional
        Recovery rate parameter, representing the rate at which infected individuals recover. If None, must be initialized separately.
    cure_rate : float, optional
        Natural immunity rate parameter, representing the rate at which individuals (across S, E, I, R compartments) return to susceptible due to loss of immunity. If None, must be initialized separately.
    latency : float, optional
        Latency rate parameter, representing the rate at which exposed individuals become infected. If None, must be initialized separately.
    population : int, optional
        Total population considered in the model. If None, the sum of the initial conditions (susceptible, exposed, infected, recovered) is used as the total population.

    Attributes
    ----------
    beta : torch.nn.Linear
        Linear layer with no bias to dynamically model the infection rate.
    gamma : torch.nn.Linear
        Linear layer with no bias to dynamically model the recovery rate.
    mu : torch.nn.Linear
        Linear layer with no bias to model the natural immunity rate.
    a : torch.nn.Linear
        Linear layer with no bias to model the latency rate.

    Returns
    -------
    torch.Tensor
        A tensor of shape (horizon, 4), representing the predicted number of susceptible, exposed, infected, and recovered individuals at each timestep.
        Each row corresponds to a timestep, with the columns representing the counts of susceptible, exposed, infected, and recovered individuals respectively.

    """
    def __init__(self, horizon=None, infection_rate = None, recovery_rate = None, cure_rate = None, latency = None, population = None):
        super(SEIR, self).__init__()
        self.pop = population

        self.beta = nn.Linear(1, 1, bias = False)
        self.gamma = nn.Linear(1, 1, bias = False)
        self.mu = nn.Linear(1, 1, bias = False)
        self.a = nn.Linear(1, 1, bias = False)

        if infection_rate is not None:
            self.beta.weight.data = torch.FloatTensor([infection_rate])

        if recovery_rate is not None:
            self.gamma.weight.data = torch.FloatTensor([recovery_rate])

        if cure_rate is not None:
            self.mu.weight.data = torch.FloatTensor([cure_rate])

        if latency is not None:
            self.a.weight.data = torch.FloatTensor([latency])
        
    def forward(self, x, steps=1):
        """
        Parameters
        ----------
        x : torch.Tensor
            The initial condition tensor for the model, representing the initial numbers of susceptible (S), exposed (E), 
            infected (I), and recovered (R) individuals. Expected shape is (4,), where elements correspond to S, E, I, and R counts.
        steps : int, optional
            Number of future time steps to simulate. If `horizon` is specified during initialization and not None, 
            it overrides this parameter. Default is 1 if `horizon` is None.

        Returns
        -------
        torch.Tensor
            A tensor of shape (steps, 4), representing the predicted number of susceptible, exposed, infected, and recovered 
            individuals at each timestep. Each row corresponds to a timestep, with columns representing the counts of 
            susceptible, exposed, infected, and recovered individuals respectively.
        """
        if self.pop is not None:
            pop = self.pop
        else:
            pop = x.sum()
        if self.horizon is not None:
            steps = self.horizon

        assert len(x.shape) == 1 and x.shape[0] == 4

        output = torch.zeros(steps* 4, dtype=torch.float, requires_grad = False).reshape(steps, 4)
        output.data[0] = x.data

        for i in range(1, steps):
            output.data[i][0] = output.data[i-1][0] - self.beta((output.data[i-1][0] * output.data[i-1][1]).unsqueeze(0))/pop + self.mu(pop - output.data[i-1][0]) # S
            output.data[i][1] = output.data[i-1][1] + self.beta((output.data[i-1][0] * output.data[i-1][1]).unsqueeze(0))/pop - self.mu(output.data[i-1][1]) - self.a(output.data[i-1][1])# E
            output.data[i][2] = output.data[i-1][2] + self.a(output.data[i-1][1]) - self.gamma(output.data[i-1][2]) - self.mu(output.data[i-1][2])# I
            output.data[i][3] = output.data[i-1][3] + self.gamma(output.data[i-1][2]) - self.mu(output.data[i-1][3])# R

        return output



import torch
import torch.nn as nn

class NetSIR(nn.Module):
    """
    Network-based SIR (Susceptible-Infected-Recovered) 

    Parameters
    ----------
    num_nodes : int, optional
        Number of nodes in the graph representing individuals or groups. Default: None.
    horizon : int, optional
        Number of future time steps to simulate. If None, a single step is simulated unless overridden in the forward method.
    infection_rate : float, optional
        Initial infection rate parameter, representing the rate at which susceptible individuals become infected. Default: 0.01.
    recovery_rate : float, optional
        Initial recovery rate parameter, representing the rate at which infected individuals recover. Default: 0.038.
    population : int, optional
        Total population considered in the model. If None, the sum of the initial conditions (susceptible, infected, recovered) is used as the total population.

    Returns
    -------
    torch.Tensor
        A tensor of shape (time_step, num_nodes, 3), representing the predicted number of susceptible, infected, and recovered individuals at each timestep for each node.
        Each row corresponds to a timestep, with the columns representing the susceptible, infected, and recovered counts respectively for each node.
    """
    def __init__(self, num_nodes=None, horizon=None, infection_rate=0.01, recovery_rate=0.038, population=None):
        super(NetSIR, self).__init__()
        self.pop = population
        self.horizon = horizon
        self.num_nodes = num_nodes

        self.beta = torch.abs(torch.rand(num_nodes))
        self.gamma = torch.abs(torch.rand(num_nodes))

        if infection_rate is not None: 
            new_weights = torch.zeros_like(self.beta.data) + torch.FloatTensor([infection_rate])
            self.beta.data = new_weights

        if recovery_rate is not None:
            new_weights = torch.zeros_like(self.gamma.data) + torch.FloatTensor([recovery_rate])
            self.gamma.data = new_weights
        
        self.beta = nn.Parameter(self.beta)
        self.gamma = nn.Parameter(self.gamma)
        
    def forward(self, x, adj, steps=1):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input features tensor with shape (n_nodes, one-hot encoding of states).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (time_step, n_nodes, probability of states),
            representing the predicted values for each node over the specified output timesteps.
        """
        if self.pop is not None:
            pop = self.pop
        else:
            pop = x.sum()
        if self.horizon is not None:
            steps = self.horizon

        # rescale = nn.Softmax(dim=0)

        output = torch.zeros(self.num_nodes*steps*3, dtype=torch.float, requires_grad = False).reshape(steps, self.num_nodes, 3)
        output.data[0] = x.data
        adj = adj.float()

        for i in range(1, steps):
            new_cases = self.beta*(output.data[i-1,:,0]*(adj @ output.data[i-1,:,1])).unsqueeze(0)
            new_recovery = self.gamma*output.data[i-1,:,1]

            output.data[i,:,0] = output.data[i-1,:,0] - new_cases
            output.data[i,:,1] = output.data[i-1,:,1] + new_cases - new_recovery
            output.data[i,:,2] = output.data[i-1,:,2] + new_recovery
            
            # output.data[i,:] = rescale(output.data[i,:])
        
        return output



# this code is adopted from https://github.com/amy-deng/colagnn/blob/master/src/models.py

import torch
import torch.nn as nn
from torch.nn import Parameter
import torch.nn.functional as F
import math

from .base import BaseModel

class GraphConvLayer(nn.Module):
    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.Tensor(in_features, out_features))
        nn.init.xavier_uniform_(self.weight)

        if bias:
            self.bias = Parameter(torch.Tensor(out_features))
            stdv = 1. / math.sqrt(self.bias.size(0))
            self.bias.data.uniform_(-stdv, stdv)
        else:
            self.register_parameter('bias', None)

    def forward(self, feature, adj):
        support = torch.matmul(feature, self.weight)
        output = torch.matmul(adj, support)

        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')' 


class ColaGNN(BaseModel):  
    """
    Convolutional-Layer Graph Neural Network (ColaGNN)

    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features per node per timestep.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    nhid : int, optional
        Number of hidden units in the RNN and GNN layers. Default: 32.
    n_channels : int, optional
        Number of channels for the convolution layers. Default: 1.
    rnn_model : str, optional
        Type of RNN model to use ('LSTM', 'GRU', 'RNN'). Default: 'GRU'.
    n_layer : int, optional
        Number of layers in the RNN model. Default: 1.
    bidirect : bool, optional
        Whether the RNN layers are bidirectional. Default: False.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output, num_nodes), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.
    """
    def __init__(self, 
                num_nodes,
                num_features,
                num_timesteps_input,
                num_timesteps_output,
                nhid=32,
                n_channels=1,
                rnn_model = 'GRU',
                n_layer = 1,
                bidirect = False,
                dropout = 0.5,
                device='cpu'): 
        
        super().__init__()
        self.device = device
        self.x_h = num_features 
        self.m = num_nodes
        self.w = num_timesteps_input
        self.h = num_timesteps_output

        self.dropout = dropout
        self.n_hidden = nhid
        half_hid = int(self.n_hidden/2)
        self.V = Parameter(torch.Tensor(half_hid))
        self.bv = Parameter(torch.Tensor(1))
        self.W1 = Parameter(torch.Tensor(half_hid, self.n_hidden))
        self.b1 = Parameter(torch.Tensor(half_hid))
        self.W2 = Parameter(torch.Tensor(half_hid, self.n_hidden))
        self.act = F.elu 
        self.Wb = Parameter(torch.Tensor(self.m,self.m))
        self.wb = Parameter(torch.Tensor(1))
        self.k = n_channels
        self.conv = nn.Conv1d(self.x_h, self.k, self.w)
        long_kernal = self.w//2
        self.conv_long = nn.Conv1d(self.x_h, self.k, long_kernal, dilation=2)
        long_out = self.w-2*(long_kernal-1)
        self.n_spatial = 10  

        self.conv1 = GraphConvLayer((1+long_out)*self.k, self.n_hidden) # self.k
        self.conv2 = GraphConvLayer(self.n_hidden, self.n_spatial)
 
        if rnn_model == 'LSTM':
            self.rnn = nn.LSTM( input_size=self.x_h, hidden_size=self.n_hidden, num_layers=n_layer, dropout=dropout, batch_first=True, bidirectional=bidirect)
        elif rnn_model == 'GRU':
            self.rnn = nn.GRU( input_size=self.x_h, hidden_size=self.n_hidden, num_layers=n_layer, dropout=dropout, batch_first=True, bidirectional=bidirect)
        elif rnn_model == 'RNN':
            self.rnn = nn.RNN( input_size=self.x_h, hidden_size=self.n_hidden, num_layers=n_layer, dropout=dropout, batch_first=True, bidirectional=bidirect)
        else:
            raise LookupError (' only support LSTM, GRU and RNN')

        hidden_size = (int(bidirect) + 1) * self.n_hidden
        self.out = nn.Linear(hidden_size + self.n_spatial, self.h)  

        self.residual_window = 0
        self.ratio = 1.0
        if (self.residual_window > 0):
            self.residual_window = min(self.residual_window, self.w)
            self.residual = nn.Linear(self.residual_window, 1) 
        self.init_weights()
     
    def init_weights(self):
        for p in self.parameters():
            if p.data.ndimension() >= 2:
                nn.init.xavier_uniform_(p.data) # best
            else:
                stdv = 1. / math.sqrt(p.size(0))
                p.data.uniform_(-stdv, stdv)

    def forward(self, x, adj, states=None, dynamic_adj=None):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input features tensor with shape (batch_size, num_timesteps_input, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_timesteps_output, num_nodes),
            representing the predicted values for each node over the specified output timesteps.
        """
        b, _,_,_ = x.size()
        orig_x = x 

        x = x.transpose(2, 1).contiguous().flatten(0,1)
        r_out, hc = self.rnn(x, None)
        last_hid = r_out[:,-1,:]
        last_hid = last_hid.view(-1,self.m, self.n_hidden)
        out_temporal = last_hid  # [b, m, 20]

        hid_rpt_m = last_hid.repeat(1,self.m,1).view(b,self.m,self.m,self.n_hidden) # b,m,m,w continuous m
        hid_rpt_w = last_hid.repeat(1,1,self.m).view(b,self.m,self.m,self.n_hidden) # b,m,m,w continuous w one window data
        a_mx = self.act( hid_rpt_m @ self.W1.t()  + hid_rpt_w @ self.W2.t() + self.b1 ) @ self.V + self.bv # row, all states influence one state 
        a_mx = F.normalize(a_mx, p=2, dim=1, eps=1e-12, out=None)

        r_l = []
        r_long_l = []
        h_mids = orig_x
        for i in range(self.m):
            h_tmp = h_mids[:,:,i].transpose(2,1).contiguous() 
            r = self.conv(h_tmp) # [32, 10/k, 1]
            r_long = self.conv_long(h_tmp)
            r_l.append(r)
            r_long_l.append(r_long)
        r_l = torch.stack(r_l,dim=1)
        r_long_l = torch.stack(r_long_l,dim=1)
        r_l = torch.cat((r_l,r_long_l),-1)
        r_l = r_l.view(r_l.size(0),r_l.size(1),-1)
        r_l = torch.relu(r_l)
        adjs = adj.repeat(b,1)
        adjs = adjs.view(b,self.m,self.m)
        c = torch.sigmoid(a_mx @ self.Wb + self.wb)
        a_mx = adjs * c + a_mx * (1-c) 
        adj = a_mx 
        
        x = r_l  
        x = F.relu(self.conv1(x, adj))
        x = F.dropout(x, self.dropout, training=self.training)

        out_spatial = F.relu(self.conv2(x, adj))
        out = torch.cat((out_spatial, out_temporal),dim=-1)
        out = self.out(out)
        out = torch.squeeze(out)

        if (self.residual_window > 0):
            z = orig_x[:, -self.residual_window:, :]; #Step backward # [batch, res_window, m]
            z = z.permute(0,2,1).contiguous().view(-1, self.residual_window); #[batch*m, res_window]
            z = self.residual(z); #[batch*m, 1]
            z = z.view(-1,self.m); #[batch, m]
            out = out * self.ratio + z; #[batch, m]

        return out.view(-1, self.h, self.m)
    
    def initialize(self):
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()



import torch

class BaseTask:
    def __init__(self, prototype, model, dataset, lookback, horizon, device='cpu'):
        self.model = model
        self.prototype = prototype
        self.dataset = dataset
        self.device = device
        self.lookback = lookback
        self.horizon = horizon
    
    def load_model(self, model):
        pass

    def load_dataset(self, dataset):
        pass

    def print_dataset(self, dataset = None):
        pass



# this code is adopted from https://github.com/nnzhan/Graph-WaveNet/blob/master/model.py

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import sys

from .base import BaseModel

class nconv(nn.Module):
    def __init__(self):
        super(nconv,self).__init__()

    def forward(self,x, A):
        x = torch.einsum('ncvl,vw->ncwl',(x,A))
        return x.contiguous()

class linear(nn.Module):
    def __init__(self,c_in,c_out):
        super(linear,self).__init__()
        self.mlp = torch.nn.Conv2d(c_in, c_out, kernel_size=(1, 1), padding=(0,0), stride=(1,1), bias=True)

    def forward(self,x):
        return self.mlp(x)

class gcn(nn.Module):
    def __init__(self,c_in,c_out,dropout,support_len=3,order=2):
        super(gcn,self).__init__()
        self.nconv = nconv()
        c_in = (order*support_len+1)*c_in
        self.mlp = linear(c_in,c_out)
        self.dropout = dropout
        self.order = order

    def forward(self,x,support):
        out = [x]
        for a in support:
            x1 = self.nconv(x,a)
            out.append(x1)
            for k in range(2, self.order + 1):
                x2 = self.nconv(x1,a)
                out.append(x2)
                x1 = x2
        h = torch.cat(out,dim=1)
        h = self.mlp(h)
        h = F.dropout(h, self.dropout, training=self.training)
        return h


class GraphWaveNet(BaseModel):
    """
    Graph Convolutional Wave Network (GraphWaveNet)

    Parameters
    ----------
    device : str
        The device (cpu or gpu) on which the model will be run.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.3.
    gcn_bool : bool, optional
        Indicates whether to include graph convolution layers. Default: True.
    addaptadj : bool, optional
        Indicates whether to include an adaptive adjacency matrix. Default: True.
    aptinit : tensor, optional
        Initial tensor for adaptive adjacency matrix. Default: None.
    num_timesteps_input : int
        Number of input timesteps per node.
    num_timesteps_output : int
        Number of output timesteps per node.
    residual_channels : int
        Number of channels in residual layers. Default: 32.
    dilation_channels : int
        Number of channels in dilation layers. Default: 32.
    skip_channels : int
        Number of channels in skip connection layers. Default: 256.
    end_channels : int
        Number of channels in the final convolution layers. Default: 512.
    kernel_size : int
        Kernel size for the convolution layers. Default: 2.
    blocks : int
        Number of blocks in the WaveNet architecture. Default: 4.
    nlayers : int
        Number of layers in each block. Default: 2.
    adj_m : tensor
        Initial adjacency matrix if static graph structure is used. Default: None.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_nodes, output_dim), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.
    """
    def __init__(self, device="cpu", dropout=0.3, gcn_bool=True, addaptadj=True, aptinit=None, num_timesteps_input=2,num_timesteps_output=12,
                 residual_channels=32,dilation_channels=32,skip_channels=256,end_channels=512,kernel_size=2,blocks=4,nlayers=2, adj_m=None):
        super(GraphWaveNet, self).__init__()

        num_nodes = adj_m.shape[0]
        self.device = device
        self.dropout = dropout
        self.blocks = blocks
        self.layers = nlayers
        self.gcn_bool = gcn_bool
        self.addaptadj = addaptadj

        self.filter_convs = nn.ModuleList()
        self.gate_convs = nn.ModuleList()
        self.residual_convs = nn.ModuleList()
        self.skip_convs = nn.ModuleList()
        self.bn = nn.ModuleList()
        self.gconv = nn.ModuleList()

        self.start_conv = nn.Conv2d(in_channels=num_timesteps_input,
                                    out_channels=residual_channels,
                                    kernel_size=(1,1))
        if adj_m is not None:
            supports = [i.clone().detach().to(device).unsqueeze(0) for i in adj_m] #supports
        else:
            supports = None
        self.supports = supports
        
        
        receptive_field = 1

        self.supports_len = 0
        if supports is not None:
            self.supports_len += len(supports)

        if gcn_bool and addaptadj:
            if aptinit is None:
                if supports is None:
                    self.supports = []
                self.nodevec1 = nn.Parameter(torch.randn(num_nodes, 10).to(device), requires_grad=True).to(device)
                self.nodevec2 = nn.Parameter(torch.randn(10, num_nodes).to(device), requires_grad=True).to(device)
                self.supports_len +=1
            else:
                if supports is None:
                    self.supports = []
                m, p, n = torch.svd(aptinit)
                initemb1 = torch.mm(m[:, :10], torch.diag(p[:10] ** 0.5))
                initemb2 = torch.mm(torch.diag(p[:10] ** 0.5), n[:, :10].t())
                self.nodevec1 = nn.Parameter(initemb1, requires_grad=True).to(device)
                self.nodevec2 = nn.Parameter(initemb2, requires_grad=True).to(device)
                self.supports_len += 1




        for b in range(blocks):
            additional_scope = kernel_size - 1
            new_dilation = 1
            for i in range(nlayers):
                # dilated convolutions
                self.filter_convs.append(nn.Conv2d(in_channels=residual_channels,
                                                   out_channels=dilation_channels,
                                                   kernel_size=(1,kernel_size),dilation=new_dilation))

                self.gate_convs.append(nn.Conv2d(in_channels=residual_channels,
                                                 out_channels=dilation_channels,
                                                 kernel_size=(1, kernel_size), dilation=new_dilation))

                # 1x1 convolution for residual connection
                self.residual_convs.append(nn.Conv2d(in_channels=dilation_channels,
                                                     out_channels=residual_channels,
                                                     kernel_size=(1, 1)))

                # 1x1 convolution for skip connection
                self.skip_convs.append(nn.Conv2d(in_channels=dilation_channels,
                                                 out_channels=skip_channels,
                                                 kernel_size=(1, 1)))
                self.bn.append(nn.BatchNorm2d(residual_channels))
                new_dilation *=2
                receptive_field += additional_scope
                additional_scope *= 2
                if self.gcn_bool:
                    self.gconv.append(gcn(dilation_channels,residual_channels,dropout,support_len=self.supports_len))



        self.end_conv_1 = nn.Conv2d(in_channels=skip_channels,
                                  out_channels=end_channels,
                                  kernel_size=(1,1),
                                  bias=True)

        self.end_conv_2 = nn.Conv2d(in_channels=end_channels,
                                    out_channels=num_timesteps_output,
                                    kernel_size=(1,1),
                                    bias=True)

        self.receptive_field = receptive_field



    def forward(self, X_batch, graph, X_states, batch_graph):
        """
        Parameters
        ----------
        X_batch : torch.Tensor
            Input features tensor with shape (batch_size, num_timesteps_input, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_timesteps_output, num_nodes),
            representing the predicted values for each node over the specified output timesteps.
        """
        #print(input)
        input = torch.permute(X_batch, (0, 2, 1, 3))
        input = input.transpose(1, 3)

        in_len = input.size(3)
        if in_len<self.receptive_field:
            x = nn.functional.pad(input,(self.receptive_field-in_len,0,0,0))
        else:
            x = input
        x = self.start_conv(x)
        skip = 0

        # calculate the current adaptive adj matrix once per iteration
        new_supports = None
        if self.gcn_bool and self.addaptadj and self.supports is not None:
            adp = F.softmax(F.relu(torch.mm(self.nodevec1, self.nodevec2)), dim=1)
            new_supports = self.supports + [adp]

        # WaveNet layers
        for i in range(self.blocks * self.layers):

            #            |----------------------------------------|     *residual*
            #            |                                        |
            #            |    |-- conv -- tanh --|                |
            # -> dilate -|----|                  * ----|-- 1x1 -- + -->	*input*
            #                 |-- conv -- sigm --|     |
            #                                         1x1
            #                                          |
            # ---------------------------------------> + ------------->	*skip*

            #(dilation, init_dilation) = self.dilations[i]
            
            #residual = dilation_func(x, dilation, init_dilation, i)
            residual = x
            # dilated convolution
            filter = self.filter_convs[i](residual)
            filter = torch.tanh(filter)
            gate = self.gate_convs[i](residual)
            gate = torch.sigmoid(gate)
            x = filter * gate
            
            # parametrized skip connection

            s = x
            s = self.skip_convs[i](s)
            try:
                skip = skip[:, :, :,  -s.size(3):]
            except:
                skip = 0
            skip = s + skip


            if self.gcn_bool and self.supports is not None:
                if self.addaptadj:
                    x = self.gconv[i](x, new_supports)
                else:
                    x = self.gconv[i](x,self.supports)
            else:
                x = self.residual_convs[i](x)

            x = x + residual[:, :, :, -x.size(3):]


            x = self.bn[i](x)

        x = F.relu(skip)
        x = F.relu(self.end_conv_1(x))
        x = self.end_conv_2(x)
        
        x = torch.permute(x, (0, 2, 1, 3)).squeeze()
        return x


    def initialize(self):
        pass



# # this code is adopted from 

# import torch
# import torch.nn as nn
# from torch_geometric.nn.inits import glorot, zeros
# from typing import Union, Tuple
# from torch_geometric.typing import OptPairTensor, Adj, Size

# from torch import Tensor, randn
# from torch.nn import Linear, Parameter
# import torch.nn.functional as F
# from torch_sparse import SparseTensor, matmul
# from torch_geometric.nn.conv import MessagePassing
# from torch import matmul

# from .base import BaseModel


# class WeightedSAGEConv(MessagePassing):
#     r"""The GraphSAGE operator from the `"Inductive Representation Learning on
#     Large Graphs" <https://arxiv.org/abs/1706.02216>`_ paper

#     .. math::
#         \mathbf{x}^{\prime}_i = \mathbf{W}_1 \mathbf{x}_i + \mathbf{W_2} \cdot
#         \mathrm{mean}_{j \in \mathcal{N(i)}} \mathbf{x}_j

#     Args:
#         in_channels (int or tuple): Size of each input sample. A tuple
#             corresponds to the sizes of source and target dimensionalities.
#         out_channels (int): Size of each output sample.
#         normalize (bool, optional): If set to :obj:`True`, output features
#             will be :math:`\ell_2`-normalized, *i.e.*,
#             :math:`\frac{\mathbf{x}^{\prime}_i}
#             {\| \mathbf{x}^{\prime}_i \|_2}`.
#             (default: :obj:`False`)
#         bias (bool, optional): If set to :obj:`False`, the layer will not learn
#             an additive bias. (default: :obj:`True`)
#         **kwargs (optional): Additional arguments of
#             :class:`torch_geometric.nn.conv.MessagePassing`.
#     """
#     def __init__(self, in_channels: Union[int, Tuple[int, int]],
#                  out_channels: int, normalize: bool = False,
#                  bias: bool = True, weighted: bool = True, edge_count: int = 423, **kwargs):  # yapf: disable
#         super(WeightedSAGEConv, self).__init__(aggr='mean', **kwargs)

#         self.in_channels = in_channels
#         self.out_channels = out_channels
#         self.normalize = normalize
#         self.weighted = weighted

#         if isinstance(in_channels, int):
#             in_channels = (in_channels, in_channels)

#         self.lin_l = Linear(in_channels[0], out_channels, bias=bias)
#         self.lin_r = Linear(in_channels[1], out_channels, bias=False)

#         self.lin_m = Linear(in_channels[0], in_channels[0], bias=bias)
#         self.lin_ew = Linear(1, 1, bias=bias)

#         self.lin_edge_feature_weighting = Linear(3, 1, bias=bias)

#         self.edge_attr = Parameter(randn(edge_count, 1))

#         self.reset_parameters()

#     def reset_parameters(self):
#         self.lin_l.reset_parameters()
#         self.lin_r.reset_parameters()

#     def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj, edge_weight: Tensor = None,
#                 size: Size = None) -> Tensor:

#         if isinstance(x, Tensor):
#             x: OptPairTensor = (x, x)

#         # propagate_type: (x: OptPairTensor)
#         out = self.propagate(edge_index=edge_index, size=size)
#         out = self.lin_l(out)

#         x_r = x[1]
#         if x_r is not None:
#             out += self.lin_r(x_r)

#         if self.normalize:
#             out = F.normalize(out, p=2., dim=-1)

#         return out

#     def message(self, x_i: Tensor, x_j: Tensor, edge_weight) -> Tensor:
#         # reduced_weight = self.lin_edge_feature_weighting(edge_weight)
#         return x_i * edge_weight # self.lin_edge_feature_weighting(edge_weight)
#         if not self.weighted:
#             return x_j
#         out = self.lin_m(x_j - x_i)
#         weight = self.lin_ew(self.edge_attr)
#         return out if self.edge_attr is None else out * weight

#     def message_and_aggregate(self, adj_t: SparseTensor,
#                               x: OptPairTensor) -> Tensor:
#         adj_t = adj_t.set_value(None, layout=None)
#         return matmul(adj_t, x[0], reduce=self.aggr)

#     def __repr__(self):
#         return '{}({}, {})'.format(self.__class__.__name__, self.in_channels,
#                                    self.out_channels)

# class LSTM(torch.nn.Module):
#     # This is an adaptation of torch_geometric_temporal.nn.GConvLSTM, with ChebConv replaced by the given model.
#     """
#     Args:
#         in_channels (int): Number of input features.
#         out_channels (int): Number of output features.
#         normalization (str, optional): The normalization scheme for the graph
#             Laplacian (default: :obj:`"sym"`):

#             1. :obj:`None`: No normalization
#             :math:`\mathbf{L} = \mathbf{D} - \mathbf{A}`

#             2. :obj:`"sym"`: Symmetric normalization
#             :math:`\mathbf{L} = \mathbf{I} - \mathbf{D}^{-1/2} \mathbf{A}
#             \mathbf{D}^{-1/2}`

#             3. :obj:`"rw"`: Random-walk normalization
#             :math:`\mathbf{L} = \mathbf{I} - \mathbf{D}^{-1} \mathbf{A}`

#             You need to pass :obj:`lambda_max` to the :meth:`forward` method of
#             this operator in case the normalization is non-symmetric.
#             :obj:`\lambda_max` should be a :class:`torch.Tensor` of size
#             :obj:`[num_graphs]` in a mini-batch scenario and a
#             scalar/zero-dimensional tensor when operating on single graphs.
#             You can pre-compute :obj:`lambda_max` via the
#             :class:`torch_geometric.transforms.LaplacianLambdaMax` transform.
#         bias (bool, optional): If set to :obj:`False`, the layer will not learn
#             an additive bias. (default: :obj:`True`)
#         module (torch.nn.Module, optional): The layer or set of layers used to calculate each gate.
#             Could also be a lambda function returning a torch.nn.Module when given the parameters in_channels: int, out_channels: int, and bias: bool
#     """
#     def __init__(self, in_channels: int, out_channels: int, bias: bool=True, module=WeightedSAGEConv):
#         super(LSTM, self).__init__()

#         self.in_channels = in_channels
#         self.out_channels = out_channels
#         self.bias = bias
#         self.module = module
#         self._create_parameters_and_layers()
#         self._set_parameters()


#     def _create_input_gate_parameters_and_layers(self):

#         self.conv_x_i = self.module(in_features=self.in_channels,
#                                  out_features=self.out_channels,
#                                  bias=self.bias)

#         self.conv_h_i = self.module(in_features=self.out_channels,
#                                  out_features=self.out_channels,
#                                  bias=self.bias)

#         self.w_c_i = Parameter(torch.Tensor(1, self.out_channels))
#         self.b_i = Parameter(torch.Tensor(1, self.out_channels))


#     def _create_forget_gate_parameters_and_layers(self):

#         self.conv_x_f = self.module(in_features=self.in_channels,
#                                  out_features=self.out_channels,
#                                  bias=self.bias)

#         self.conv_h_f = self.module(in_features=self.out_channels,
#                                  out_features=self.out_channels,
#                                  bias=self.bias)

#         self.w_c_f = Parameter(torch.Tensor(1, self.out_channels))
#         self.b_f = Parameter(torch.Tensor(1, self.out_channels))


#     def _create_cell_state_parameters_and_layers(self):

#         self.conv_x_c = self.module(in_features=self.in_channels,
#                                  out_features=self.out_channels,
#                                  bias=self.bias)

#         self.conv_h_c = self.module(in_features=self.out_channels,
#                                  out_features=self.out_channels,
#                                  bias=self.bias)

#         self.b_c = Parameter(torch.Tensor(1, self.out_channels))


#     def _create_output_gate_parameters_and_layers(self):

#         self.conv_x_o = self.module(in_features=self.in_channels,
#                                  out_features=self.out_channels,
#                                  bias=self.bias)

#         self.conv_h_o = self.module(in_features=self.out_channels,
#                                  out_features=self.out_channels,
#                                  bias=self.bias)

#         self.w_c_o = Parameter(torch.Tensor(1, self.out_channels))
#         self.b_o = Parameter(torch.Tensor(1, self.out_channels))


#     def _create_parameters_and_layers(self):
#         self._create_input_gate_parameters_and_layers()
#         self._create_forget_gate_parameters_and_layers()
#         self._create_cell_state_parameters_and_layers()
#         self._create_output_gate_parameters_and_layers()


#     def _set_parameters(self):
#         glorot(self.w_c_i)
#         glorot(self.w_c_f)
#         glorot(self.w_c_o)
#         zeros(self.b_i)
#         zeros(self.b_f)
#         zeros(self.b_c)
#         zeros(self.b_o)


#     def _set_hidden_state(self, X, H):
#         if H is None:
#             H = torch.zeros(X.shape[0], self.out_channels)
#         return H


#     def _set_cell_state(self, X, C):
#         if C is None:
#             C = torch.zeros(X.shape[0], self.out_channels)
#         return C


#     def _calculate_input_gate(self, X, edge_index, edge_weight, H, C):
#         I = self.conv_x_i(X, edge_index, edge_weight)
#         I = I + self.conv_h_i(H, edge_index, edge_weight)
#         I = I + (self.w_c_i*C)
#         I = I + self.b_i
#         I = torch.sigmoid(I)
#         return I


#     def _calculate_forget_gate(self, X, edge_index, edge_weight, H, C):
#         F = self.conv_x_f(X, edge_index, edge_weight)
#         F = F + self.conv_h_f(H, edge_index, edge_weight)
#         F = F + (self.w_c_f*C)
#         F = F + self.b_f
#         F = torch.sigmoid(F)
#         return F


#     def _calculate_cell_state(self, X, edge_index, edge_weight, H, C, I, F):
#         T = self.conv_x_c(X, edge_index, edge_weight)
#         T = T + self.conv_h_c(H, edge_index, edge_weight)
#         T = T + self.b_c
#         T = torch.tanh(T)
#         C = F*C + I*T
#         return C

#     def _calculate_output_gate(self, X, edge_index, edge_weight, H, C):
#         O = self.conv_x_o(X, edge_index, edge_weight)
#         O = O + self.conv_h_o(H, edge_index, edge_weight)
#         O = O + (self.w_c_o*C)
#         O = O + self.b_o
#         O = torch.sigmoid(O)
#         return O


#     def _calculate_hidden_state(self, O, C):
#         H = O * torch.tanh(C)
#         return H


#     def forward(self, X: torch.FloatTensor, edge_index: torch.LongTensor, edge_weight: torch.FloatTensor=None,
#                 H: torch.FloatTensor=None, C: torch.FloatTensor=None) -> torch.FloatTensor:
#         """
#         Making a forward pass. If edge weights are not present the forward pass
#         defaults to an unweighted graph. If the hidden state and cell state
#         matrices are not present when the forward pass is called these are
#         initialized with zeros.

#         Arg types:
#             * **X** *(PyTorch Float Tensor)* - Node features.
#             * **edge_index** *(PyTorch Long Tensor)* - Graph edge indices.
#             * **edge_weight** *(PyTorch Long Tensor, optional)* - Edge weight vector.
#             * **H** *(PyTorch Float Tensor, optional)* - Hidden state matrix for all nodes.
#             * **C** *(PyTorch Float Tensor, optional)* - Cell state matrix for all nodes.

#         Return types:
#             * **H** *(PyTorch Float Tensor)* - Hidden state matrix for all nodes.
#             * **C** *(PyTorch Float Tensor)* - Cell state matrix for all nodes.
#         """
#         H = self._set_hidden_state(X, H)
#         C = self._set_cell_state(X, C)
#         I = self._calculate_input_gate(X, edge_index, edge_weight, H, C)
#         F = self._calculate_forget_gate(X, edge_index, edge_weight, H, C)
#         C = self._calculate_cell_state(X, edge_index, edge_weight, H, C, I, F)
#         O = self._calculate_output_gate(X, edge_index, edge_weight, H, C)
#         H = self._calculate_hidden_state(O, C)
#         return H, C


# class GraphLinear(torch.nn.Linear):
#     """This is the exact same as torch.nn.Linear,
#     except that it can take edge_index, edge_attr and do nothing with them.
#     Makes it interchangeable with graph neural network modules."""

#     def forward(self, input, edge_index, edge_attr):
#         return super(GraphLinear, self).forward(input)


# class GraphLSTM(BaseModel):
#     """
#     Base class for Recurrent Neural Networks (LSTM or GRU).
#     Initialization to this class contains all variables for variation of the model.
#     Consists of one of the above RNN architectures followed by an optional GNN on the final hidden state.
#     Parameters:
#         node_features: int - number of features per node
#         output: int - length of the output vector on each node
#         dim: int - number of features of embedding for each node
#         module: torch.nn.Module - to be used in the LSTM to calculate each gate
#     """
#     def __init__(self, 
#                 num_nodes, 
#                 num_features,
#                 num_timesteps_input,
#                 num_timesteps_output, 
#                 dim=32, 
#                 module=GraphLinear, 
#                 rnn=LSTM, 
#                 gnn=WeightedSAGEConv, 
#                 gnn_2=WeightedSAGEConv, 
#                 rnn_depth=1, 
#                 name="RNN", 
#                 skip_connection=True):
#         super(GraphLSTM, self).__init__()
#         self.n_feat = num_features
#         self.n_in = num_timesteps_input
#         self.m = num_nodes
#         self.dim = dim
#         self.rnn_depth = rnn_depth
#         self.name = name
#         self.skip_connection = skip_connection

#         # Ensure that matrix multiplication sizes match up based on whether GNNs and RNN are used
#         if gnn:
#             if skip_connection:
#                 self.gnn = gnn(num_features, dim)
#             else:
#                 self.gnn = gnn(num_features, dim * 2)
#             if rnn:
#                 if skip_connection:
#                     self.recurrent = rnn(dim, dim, module=module)
#                 else:
#                     self.recurrent = rnn(dim * 2, dim * 2, module=module)
#             else:
#                 self.recurrent = None
#         else:
#             self.gnn = None
#             if rnn:
#                 self.recurrent = rnn(num_features, dim, module=module)
#             else:
#                 self.recurrent = None
#         if gnn_2:
#             if gnn:
#                 self.gnn_2 = gnn_2(dim * 2, dim * 2)
#             else:
#                 self.gnn_2 = gnn_2(dim + num_features, dim * 2)
#         else:
#             self.gnn_2 = None

#         self.lin1 = torch.nn.Linear(2 * dim, dim)
#         self.lin2 = torch.nn.Linear(dim, num_timesteps_output)
#         self.act1 = torch.nn.ReLU()
#         self.act2 = torch.nn.ReLU()

#     def forward(self, X, adj, states=None, dynamic_adj=None, h=None, c=None):
#         # Get data from snapshot
#         x, edge_index = X, adj.to_sparse()

#         # First GNN Layer
#         if self.gnn:
#             x = self.gnn(x, edge_index)
#             x = F.relu(x)

#         # Initialize hidden and cell states if None
#         current_dim = self.dim
#         if not self.skip_connection:
#             current_dim = self.dim * 2
#         if h is None:
#             h = torch.zeros(x.shape[0], current_dim)
#         if c is None:
#             c = torch.zeros(x.shape[0], current_dim)

#         # RNN Layer
#         if self.recurrent:
#             for i in range(self.rnn_depth):
#                 h, c = self.recurrent(x, edge_index, edge_attr, h, c)

#         # Skip connection from first GNN
#         if self.skip_connection:
#             x = torch.cat((x, h), 1)
#         else:
#             x = h

#         # Second GNN Layer
#         if self.gnn_2:
#             x = self.gnn_2(x, edge_index, edge_attr)

#         # Readout and activation layers
#         x = self.lin1(x)
#         # x = self.act1(x)
#         x = self.lin2(x)
#         # x = self.act2(x)

#         return x, h, c
    

#     def initialize(self):
#         for layer in self.children():
#             if hasattr(layer, 'reset_parameters'):
#                 layer.reset_parameters()


# this code is adopted from https://github.com/LeronQ/STGCN-Pytorch/blob/main/stgcn.py

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import BaseModel


class TimeBlock(nn.Module):
    """
    Neural network block that applies a temporal convolution to each node of
    a graph in isolation.
    """

    def __init__(self, in_channels, out_channels, kernel_size=3):
        """
        :param in_channels: Number of input features at each node in each time
        step.
        :param out_channels: Desired number of output channels at each node in
        each time step.
        :param kernel_size: Size of the 1D temporal kernel.
        """
        super(TimeBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, (1, kernel_size))
        self.conv2 = nn.Conv2d(in_channels, out_channels, (1, kernel_size))
        self.conv3 = nn.Conv2d(in_channels, out_channels, (1, kernel_size))

    def forward(self, X):
        """
        :param X: Input data of shape (batch_size, num_nodes, num_timesteps,
        num_features=in_channels)
        :return: Output data of shape (batch_size, num_nodes,
        num_timesteps_out, num_features_out=out_channels)
        """
        # Convert into NCHW format for pytorch to perform convolutions.
        X = X.permute(0, 3, 1, 2)
        temp = self.conv1(X) + torch.sigmoid(self.conv2(X))
        out = F.relu(temp + self.conv3(X))
        # Convert back from NCHW to NHWC
        out = out.permute(0, 2, 3, 1)
        return out
    
    def reset_parameters(self):
        self.conv1.reset_parameters()
        self.conv2.reset_parameters()
        self.conv3.reset_parameters()


class STGCNBlock(nn.Module):
    """
    Spatio-temporal graph convolutional network as described in
    https://arxiv.org/abs/1709.04875v3 by Yu et al.

    Spatio-Temporal Graph Convolutional Network (STGCN)

    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features per node per timestep.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    Attributes
    ----------
    block1 : STGCNBlock
        First STGCN block which applies spatial and temporal convolutions.
    block2 : STGCNBlock
        Second STGCN block which applies further spatial and temporal convolutions.
    last_temporal : TimeBlock
        Temporal convolution block that processes the output of the last STGCN block.
    fully : torch.nn.Linear
        Fully connected layer to reshape the output into the desired number of future time steps.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_nodes, num_timesteps_output), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.
    """
    def __init__(self, in_channels, spatial_channels, out_channels,
                 num_nodes):
        """
        :param in_channels: Number of input features at each node in each time
        step.
        :param spatial_channels: Number of output channels of the graph
        convolutional, spatial sub-block.
        :param out_channels: Desired number of output features at each node in
        each time step.
        :param num_nodes: Number of nodes in the graph.
        """
        super(STGCNBlock, self).__init__()
        self.temporal1 = TimeBlock(in_channels=in_channels,
                                   out_channels=out_channels)
        self.Theta1 = nn.Parameter(torch.FloatTensor(out_channels,
                                                     spatial_channels))
        self.temporal2 = TimeBlock(in_channels=spatial_channels,
                                   out_channels=out_channels)
        self.batch_norm = nn.BatchNorm2d(num_nodes)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.Theta1.shape[1])
        self.Theta1.data.uniform_(-stdv, stdv)

        self.temporal1.reset_parameters()
        self.temporal2.reset_parameters()
        self.batch_norm.reset_parameters()


    def forward(self, X, A_hat):
        """
        :param X: Input data of shape (batch_size, num_nodes, num_timesteps,
        num_features=in_channels).
        :param A_hat: Normalized adjacency matrix.
        :return: Output data of shape (batch_size, num_nodes,
        num_timesteps_out, num_features=out_channels).
        """
        t = self.temporal1(X)
        lfs = torch.einsum("ij,jklm->kilm", [A_hat, t.permute(1, 0, 2, 3)])
        # t2 = F.relu(torch.einsum("ijkl,lp->ijkp", [lfs, self.Theta1]))
        t2 = F.relu(torch.matmul(lfs, self.Theta1))
        t3 = self.temporal2(t2)
        return self.batch_norm(t3)
        # return t3


class STGCN(BaseModel):
    """
    Spatio-temporal graph convolutional network as described in
    https://arxiv.org/abs/1709.04875v3 by Yu et al.
    Input should have shape (batch_size, num_nodes, num_input_time_steps,
    num_features).
    """

    def __init__(self, num_nodes, num_features, num_timesteps_input,
                 num_timesteps_output, device = 'cpu'):
        """
        :param num_nodes: Number of nodes in the graph.
        :param num_features: Number of features at each node in each time step.
        :param num_timesteps_input: Number of past time steps fed into the
        network.
        :param num_timesteps_output: Desired number of future time steps
        output by the network.
        """
        super(STGCN, self).__init__(device=device)
        self.block1 = STGCNBlock(in_channels=num_features, out_channels=64,
                                 spatial_channels=16, num_nodes=num_nodes)
        self.block2 = STGCNBlock(in_channels=64, out_channels=64,
                                 spatial_channels=16, num_nodes=num_nodes)
        self.last_temporal = TimeBlock(in_channels=64, out_channels=64)
        self.fully = nn.Linear((num_timesteps_input - 2 * 5) * 64,
                               num_timesteps_output)


    def forward(self, X, adj, states=None, dynamic_adj=None, **kargs):
        """
        Parameters
        ----------
        X : torch.Tensor
            Shape (batch_size, num_nodes, num_timesteps_input, num_features)
        adj : torch.Tensor
            Shape (num_nodes, num_nodes)

        Returns
        -------
        torch.Tensor
            Output shape (batch_size, num_timesteps_output, num_nodes)
        """
        out1 = self.block1(X, adj)
        out2 = self.block2(out1, adj)
        out3 = self.last_temporal(out2)
        out4 = self.fully(out3.reshape((out3.shape[0], out3.shape[1], -1)))
        return out4
    
    def initialize(self):
        self.block1.reset_parameters()
        self.block2.reset_parameters()
        self.last_temporal.reset_parameters()
        self.fully.reset_parameters()



import torch
import os
import matplotlib.pyplot as plt
#os.sys.path.append("../Epilearn")
os.sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from epilearn.models.SpatialTemporal.STGCN import STGCN
from epilearn.models.Spatial.GCN import GCN
from epilearn.models.Spatial.GAT import GAT
from epilearn.models.Spatial.SAGE import SAGE
from epilearn.models.Spatial.GIN import GIN
from epilearn.models.SpatialTemporal.DCRNN import DCRNN
from epilearn.models.SpatialTemporal.GraphWaveNet import GraphWaveNet
from epilearn.data import UniversalDataset
from epilearn.utils import utils, metrics, transforms

import torch_geometric


# initial settings
#device = torch.device('cpu')
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print(device)

torch.manual_seed(7)

lookback = 13 # inputs size
horizon = 1 # predicts size

epochs = 50 # training epochs
batch_size = 8 # training batch size

permute = True

# load toy dataset
dataset = UniversalDataset()
dataset.load_toy_dataset()

# preprocessing
features, mean, std = utils.normalize(dataset.x)
adj_norm = utils.normalize_adj(dataset.graph)
adj_dynamic_norm = utils.normalize_adj(dataset.dynamic_graph)

features = features.to(device)
adj_norm = adj_norm.to(device)
adj_dynamic_norm = adj_dynamic_norm.to(device)

# prepare datasets
train_rate = 0.6 
val_rate = 0.2

target_feat_idx = None
target_idx = None

split_line1 = int(features.shape[0] * train_rate)
split_line2 = int(features.shape[0] * (train_rate + val_rate))

train_original_data = features[:split_line1, :, :]
val_original_data = features[split_line1:split_line2, :, :]
test_original_data = features[split_line2:, :, :]

train_original_states = dataset.states[:split_line1, :, :]
val_original_states = dataset.states[split_line1:split_line2, :, :]
test_original_states = dataset.states[split_line2:, :, :]

train_input, train_target, train_states, train_adj = dataset.generate_dataset(X = train_original_data, Y = train_original_data[..., 0], states = train_original_states, dynamic_adj = adj_dynamic_norm, lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)
val_input, val_target, val_states, val_adj = dataset.generate_dataset(X = val_original_data, Y = val_original_data[..., 0], states = val_original_states, dynamic_adj = adj_dynamic_norm, lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)
test_input, test_target, test_states, test_adj = dataset.generate_dataset(X = test_original_data, Y = test_original_data[..., 0], states = test_original_states, dynamic_adj = adj_dynamic_norm, lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)

rows, cols = torch.where(adj_norm != 0)
weights = adj_norm[rows, cols]

edge_index = torch.stack([rows.long(), cols.long()], dim=0).to(device)
edge_weight = weights.clone().detach().to(device)


print("original: ", train_input.shape)
data = {"features": train_input}
'''transform_lists = torch.nn.Sequential(transforms.convert_to_frequency(ftype="fft"), transforms.test_seq())
#transform_list = [transforms.convert_to_frequency()]
transform_list = torch.nn.Sequential(transforms.learnable_time_embedding(timesteps=13,embedding_dim=10))
#transform_list = torch.nn.Sequential(transforms.add_time_embedding(embedding_dim=10))
transform = transforms.Compose(transform_list, device)'''

transform_list = {"features": [transforms.add_time_embedding(embedding_dim=10)]}
transform = transforms.Compose(transform_list, device)

data = transform(data)
#val_input = transform(val_input)

print("transform: ", data["features"].shape)


#train_dataset = UniversalDataset(x=train_input[..., 0,:], y=train_target, graph=adj_norm, edge_index=edge_index, edge_weight=edge_weight)
#val_dataset = UniversalDataset(x=val_input[..., 0,:], y=val_target, graph=adj_norm, edge_index=edge_index, edge_weight=edge_weight)



model = GCN(num_features=train_input.shape[3],
        hidden_dim=16,
        num_classes=horizon,
        nlayers=2, with_bn=True,
        dropout=0.3, device=device)

'''model = GAT(num_features=train_input.shape[3],
        hidden_dim=16,
        num_classes=horizon,
        nlayers=2, with_bn=True, nheads=[2,4], concat=True,
        dropout=0.3, device=device)'''

'''model = SAGE(num_features=train_input.shape[2]*train_input.shape[3],
        hidden_dim=16,
        num_classes=horizon,
        nlayers=1, with_bn=True, aggr=torch_geometric.nn.GRUAggregation,
        dropout=0.3, device=device)'''

'''model = GIN(num_features=train_input.shape[2]*train_input.shape[3],
        hidden_dim=16,
        num_classes=horizon,
        nlayers=2, 
        dropout=0.3, device=device)'''


'''model = DCRNN(num_features=train_input.shape[3],
              num_timesteps_input=lookback,
              num_timesteps_output=horizon,
              num_classes=1,
              max_diffusion_step=2, 
              filter_type="laplacian",
              num_rnn_layers=1, 
              rnn_units=1,
              nonlinearity="tanh",
              dropout=0.5,
              device=device)'''


'''model = GraphWaveNet(num_timesteps_input=train_input.shape[3], num_timesteps_output=horizon, 
                     adj_m=adj_norm, gcn_bool=True, 
                     addaptadj=True, aptinit=None, 
                     blocks=4, nlayers=2,
                     residual_channels=32, dilation_channels=32, 
                     skip_channels=32 * 8, end_channels=32 * 16,dropout=0.3, device=device)'''

model = model.to(device)


model.fit(
        train_input=train_input[..., 0,:], 
        train_target=train_target, 
        train_states=None, 
        train_graph=adj_norm, 
        train_dynamic_graph=None,
        val_input=val_input[..., 0,:], 
        val_target=val_target,
        val_states=None, 
        val_graph=adj_norm, 
        val_dynamic_graph=None,
        loss='mse', 
        epochs=100, 
        batch_size=10,
        lr=1e-3, 
        weight_decay=1e-3,
        initialize=True, 
        verbose=True, 
        patience=10, 
        shuffle=False,
        )
        
'''model.fit(
        train = train_dataset, 
        val_dataset = val_dataset, 
        verbose = True, 
        batch_size = batch_size,
        lr=1e-3,
        shuffle=False,
        patience=5,
        epochs = epochs)
'''


# evaluate
#output_dim=[47, horizon]
#test_dataset = UniversalDataset(x=test_input[..., 0,:], edge_index=edge_index)
#out = model.predict(dataset=test_dataset, batch_size=batch_size, device=device, output_dim=output_dim)
out = model.predict(feature=test_input[..., 0,:], 
                    graph=adj_norm, 
                    states=None, 
                    dynamic_graph=None, 
                    batch_size=1, 
                    device = device, 
                    shuffle=False)
#out = model.predict(feature=test_input, graph=adj_norm, states=None, dynamic_graph=None)
print(out.shape)

preds = out.detach().cpu() * std[0] + mean[0]
targets = test_target.detach().cpu() * std[0] + mean[0]
# MAE
mae = metrics.get_MAE(preds, targets)
print(f"MAE: {mae.item()}")


import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GATConv, GATv2Conv
import torch

from .base import BaseModel

class GAT(BaseModel):
    """
    Graph Attention Network (GAT)
    
    Parameters
    ----------
    num_features : int
        Number of input features per node.
    hidden_dim : int
        Dimension of hidden layers.
    num_classes : int
        Number of output features per node.
    nlayers : int, optional
        Number of layers in the GAT. Default: 2.
    nheads : list of int
        Number of attention heads in each GAT layer. Length must match `nlayers`.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    with_bn : bool, optional
        Specifies whether batch normalization should be included. Default: False.
    with_bias : bool, optional
        Specifies whether to include bias parameters in the attention calculations. Default: True.
    device : torch.device
        The device (cpu or gpu) on which the model will be run.
    concat : bool, optional
        Specifies whether to concatenate the outputs of the attention heads instead of averaging them. Default: False.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_nodes, output_dim), representing the predicted values for each node over future timesteps.
    """
    def __init__(self, num_features, hidden_dim, num_classes, nlayers=2, nheads=[2, 2], dropout=0.5,
                with_bn=False, with_bias=True, device=None, concat=False):

        super(GAT, self).__init__()

        assert device is not None, "Please specify 'device'!"
        self.device = device
        
        assert len(nheads) == nlayers, "nheads should match nlayers!"

        self.layers = nn.ModuleList([])
        if with_bn:
            self.bns = nn.ModuleList()
            
        
        if concat:
            for i in range(nlayers):
                assert hidden_dim % nheads[i] == 0, "Hidden_dim should be divisible by nheads!"
                #assert num_classes % nheads[i] == 0, "Output_dim should be divisible by nheads!"
        
        if concat:
            if nlayers == 1:
                self.layers.append(GATConv(num_features, hidden_dim//nheads[0], bias=with_bias, heads=nheads[0], concat=concat))
            else:
                self.layers.append(GATConv(num_features, hidden_dim//nheads[0], bias=with_bias, heads=nheads[0], concat=concat))
                if with_bn:
                    self.bns.append(nn.BatchNorm1d(hidden_dim))
                        
                for i in range(nlayers-2):
                    self.layers.append(GATConv(hidden_dim, hidden_dim//nheads[i+1], bias=with_bias, heads=nheads[i+1], concat=concat))
                    if with_bn:
                        self.bns.append(nn.BatchNorm1d(hidden_dim))
                self.layers.append(GATConv(hidden_dim, hidden_dim//nheads[-1], bias=with_bias, heads=nheads[-1], concat=concat))
            
            
        else:
            if nlayers == 1:
                self.layers.append(GATConv(num_features, hidden_dim, bias=with_bias, heads=nheads[0], concat=concat))
            else:
                self.layers.append(GATConv(num_features, hidden_dim, bias=with_bias, heads=nheads[0], concat=concat))
                if with_bn:
                    self.bns.append(nn.BatchNorm1d(hidden_dim))
                        
                for i in range(nlayers-2):
                    self.layers.append(GATConv(hidden_dim, hidden_dim, bias=with_bias, heads=nheads[i+1], concat=concat))
                    if with_bn:
                        self.bns.append(nn.BatchNorm1d(hidden_dim))
                self.layers.append(GATConv(hidden_dim, hidden_dim, bias=with_bias, heads=nheads[-1], concat=concat))
                
            
        self.fc = nn.Linear(hidden_dim, num_classes)
            
        self.dropout = dropout
        self.with_bn = with_bn

    def forward(self, x, edge_index, edge_weight):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input features tensor with shape (num_nodes, num_features).
        edge_index : torch.Tensor
            Tensor defining the edges of the graph with shape (2, num_edges), where each column represents an edge as a pair of node indices.
        edge_weight : torch.Tensor, optional
            Edge weights with shape (num_edges,). Default is None.

        Returns
        -------
        torch.Tensor
            Output tensor of shape (num_nodes, num_classes), representing the predicted values for each node.
        """
        #x, edge_index, edge_attr = data.x, data.edge_index, data.edge_attr
        x = torch.reshape(x, (x.shape[0], -1))

        for i, layer in enumerate(self.layers):
            if edge_weight is not None:
                x = layer(x.float(), edge_index=edge_index, edge_attr=edge_weight)
            else:
                x = layer(x.float(), edge_index=edge_index)
            if i != len(self.layers) - 1:
                if self.with_bn:
                    x = self.bns[i](x)
                x = F.relu(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        return self.fc(x) #F.log_softmax(x, dim=1)
        #return F.log_softmax(self.fc(x), dim=1)

    def initialize(self):
        for m in self.layers:
            m.reset_parameters()
        if self.with_bn:
            for bn in self.bns:
                bn.reset_parameters()



import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import SAGEConv
import torch
from torch_geometric.utils import sort_edge_index

from .base import BaseModel

class SAGE(BaseModel):
    """
    Graph Sample and Aggregate (SAGE)

    Parameters
    ----------
    num_features : int
        Number of input features per node.
    hidden_dim : int
        Dimension of hidden layers.
    num_classes : int
        Number of output features per node.
    nlayers : int, optional
        Number of layers in the GraphSAGE model. Default: 2.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    with_bn : bool, optional
        Specifies whether batch normalization should be included. Default: False.
    with_bias : bool, optional
        Specifies whether to include bias parameters in the GraphSAGE layers. Default: True.
    device : str
        The device (cpu or gpu) on which the model will be run. Must be specified.
    aggr : str or callable, optional
        The aggregation function to use ('mean', 'sum', 'max', etc.), or a callable that returns a custom aggregation function. Default: 'mean'.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_nodes, output_dim), representing the predicted outcomes for each node after passing through the GraphSAGE model.
    """
    def __init__(self, num_features, hidden_dim, num_classes, nlayers=2, dropout=0.5,
                with_bn=False, with_bias=True, device=None, aggr="mean"):

        super(SAGE, self).__init__()

        assert device is not None, "Please specify 'device'!"
        self.device = device

        self.layers = nn.ModuleList([])
        if with_bn:
            self.bns = nn.ModuleList()

        if nlayers == 1:
            if callable(aggr):
                try:
                    self.layers.append(SAGEConv(num_features, hidden_dim, bias=with_bias, aggr=aggr(num_features, num_features)))
                except:
                    self.layers.append(SAGEConv(num_features, hidden_dim, bias=with_bias, aggr=aggr()))
            else:
                self.layers.append(SAGEConv(num_features, hidden_dim, bias=with_bias, aggr=aggr))
        else:
            if callable(aggr):
                try:
                    self.layers.append(SAGEConv(num_features, hidden_dim, bias=with_bias, aggr=aggr(num_features, num_features)))
                except:
                    self.layers.append(SAGEConv(num_features, hidden_dim, bias=with_bias, aggr=aggr()))
            else:
                self.layers.append(SAGEConv(num_features, hidden_dim, bias=with_bias, aggr=aggr))

            if with_bn:
                self.bns.append(nn.BatchNorm1d(hidden_dim))
            for i in range(nlayers-2):
                if callable(aggr):
                    try:
                        self.layers.append(SAGEConv(hidden_dim, hidden_dim, bias=with_bias, aggr=aggr(hidden_dim, hidden_dim)))
                    except:
                        self.layers.append(SAGEConv(hidden_dim, hidden_dim, bias=with_bias, aggr=aggr()))
                else:
                    self.layers.append(SAGEConv(hidden_dim, hidden_dim, bias=with_bias, aggr=aggr))
                
                if with_bn:
                    self.bns.append(nn.BatchNorm1d(hidden_dim))
            if callable(aggr):
                try:
                    self.layers.append(SAGEConv(hidden_dim, hidden_dim, bias=with_bias, aggr=aggr(hidden_dim, hidden_dim)))
                except:
                    self.layers.append(SAGEConv(hidden_dim, hidden_dim, bias=with_bias, aggr=aggr()))
            else:
                self.layers.append(SAGEConv(hidden_dim, hidden_dim, bias=with_bias, aggr=aggr))

        self.fc = nn.Linear(hidden_dim, num_classes)
        self.dropout = dropout
        self.with_bn = with_bn

    def forward(self, x, edge_index, edge_weight):
        """
        Parameters:
        x : torch.Tensor
            Node feature matrix with shape (batch_size, num_nodes, num_features).
        edge_index : torch.Tensor
            Edge index in COO format with shape (2, num_edges).

        Returns:
        torch.Tensor
            Output from the network with shape (batch_size * num_nodes, num_classes).
        """
        #x, edge_index, edge_weight = data.x, data.edge_index, data.edge_attr
        x = torch.reshape(x, (x.shape[0], -1))

        edge_index = sort_edge_index(edge_index, sort_by_row=False)
        for i, layer in enumerate(self.layers):
            x = layer(x, edge_index)
            if i != len(self.layers) - 1:
                if self.with_bn:
                    x = self.bns[i](x)
                x = F.relu(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        return self.fc(x) #F.log_softmax(x, dim=1)

    def initialize(self):
        for m in self.layers:
            m.reset_parameters()
        if self.with_bn:
            for bn in self.bns:
                bn.reset_parameters()



import torch
import torch.nn as nn
import torch.nn.functional as F

#——————————————————————————losses

def get_loss(loss_name = 'mse'):
    """
    Retrieves the specified loss function based on the input loss name. It supports mean squared error (MSE),
    a standardized loss (stan), an epidemic-collaboration specific loss (epi_cola), and cross-entropy loss.

    Parameters
    ----------
    loss_name : str, optional
        Name of the loss function to retrieve. Default is 'mse'.

    Returns
    -------
    callable
        The corresponding loss function as specified by loss_name.
    """
    loss_name = loss_name.lower()
    if loss_name == 'mse':
        return nn.MSELoss()
    elif loss_name == 'stan':
        return stan_loss
    elif loss_name == 'epi_cola':
        return epi_cola_loss
    elif loss_name == 'ce':
        return cross_entropy_loss


def stan_loss(output, label, scale=0.5):
    """
    Calculates a combined mean squared error loss on predicted and physically informed predicted values,
    scaled by a given factor.

    Parameters
    ----------
    output : tuple of torch.Tensor
        The predicted values and the physically informed predicted values.
    label : torch.Tensor
        The ground truth values.
    scale : float, optional
        Scaling factor for the physical informed loss component. Default: 0.5.

    Returns
    -------
    torch.Tensor
        The calculated total loss as a scalar tensor.
    """
    pred_IR, pred_phy_IR = output
    mse = nn.MSELoss()
    total_loss = mse(pred_IR, label) + scale*mse(pred_phy_IR, label)
    return total_loss

def epi_cola_loss(output, label, scale=0.5):
    """
    Calculates a combined L1 and mean squared error loss on the output and an epidemiological output,
    scaled by a given factor.

    Parameters
    ----------
    output : tuple of torch.Tensor
        The primary model output and the epidemiological model output.
    label : torch.Tensor
        The ground truth values.
    scale : float, optional
        Scaling factor for the epidemiological loss component. Default: 0.5.

    Returns
    -------
    torch.Tensor
        The calculated total loss as a scalar tensor.
    """
    output, epi_output = output
    mse = nn.MSELoss()
    total_loss = F.l1_loss(output, label) + scale*mse(epi_output, label)
    return total_loss

def cross_entropy_loss(output, label):
    """
    Computes the cross-entropy loss between the logits and labels, adjusting the label tensor to fit the logits dimensions.

    Parameters
    ----------
    output : torch.Tensor
        The logits from the model.
    label : torch.Tensor
        The ground truth labels, scaled to match the number of classes based on output dimensions.

    Returns
    -------
    torch.Tensor
        The cross-entropy loss as a scalar tensor.
    """
    label = (((label-label.min())/(label.max()-label.min()+1))*output.shape[-1]).int()
    ce = nn.CrossEntropyLoss()
    return ce(output.float().view(-1, output.shape[-1]), label.long().view(-1))


#--------------------metrics------------------
def get_MAE(pred, target):
    """
    Calculates the Mean Absolute Error (MAE) between predictions and targets.

    Parameters
    ----------
    pred : torch.Tensor
        Predicted values.
    target : torch.Tensor
        Ground truth values.

    Returns
    -------
    torch.Tensor
        The MAE value as a scalar tensor.
    """
    return torch.mean(torch.absolute(pred - target))

def get_RMSE(pred, target):
    """
    Calculates the Root Mean Squared Error (RMSE) between predictions and targets.

    Parameters
    ----------
    pred : torch.Tensor
        Predicted values.
    target : torch.Tensor
        Ground truth values.

    Returns
    -------
    torch.Tensor
        The RMSE value as a scalar tensor.
    """
    mse_loss = nn.MSELoss(reduction='mean')
    return torch.sqrt(mse_loss(pred, target))

def get_ACC(pred, target):
    """
    Calculates the accuracy of predictions by comparing them to the targets.

    Parameters
    ----------
    pred : torch.Tensor
        Predicted labels.
    target : torch.Tensor
        True labels.

    Returns
    -------
    torch.Tensor
        The accuracy as a scalar tensor.
    """
    result = pred.eq(target).sum()/len(pred)
    return result



# this code is adopted from https://github.com/deepkashiwa20/MepoGNN/blob/main/model/MepoGNN.py

import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import BaseModel

class nconv(nn.Module):
    def __init__(self):
        super(nconv,self).__init__()

    def forward(self,x, A):
        if len(A.shape) == 2:
            x = torch.einsum('vw, ncwl->ncvl', A, x)
        else:
            x = torch.einsum('nvw, ncwl->ncvl', A, x)
        return x.contiguous()


class linear(nn.Module):
    def __init__(self,c_in,c_out):
        super(linear,self).__init__()
        self.mlp = torch.nn.Conv2d(c_in, c_out, kernel_size=(1, 1), padding=(0,0), stride=(1,1), bias=True)

    def forward(self, x):
        return self.mlp(x)

class gcn(nn.Module):
    def __init__(self, c_in, c_out, dropout, support_len=2, order=1):
        super(gcn, self).__init__()
        self.nconv = nconv()
        c_in = (order*support_len)*c_in
        self.mlp = linear(c_in,c_out)
        self.dropout = dropout
        self.order = order

    def forward(self, x, support):
        out = []
        for a in support:
            x1 = self.nconv(x, a)
            out.append(x1)
            for k in range(2, self.order + 1):
                x2 = self.nconv(x1, a)
                out.append(x2)
                x1 = x2

        h = torch.cat(out,dim=1)
        h = self.mlp(h)
        h = F.dropout(h, self.dropout, training=self.training)
        return h


class stcell(nn.Module):
    def __init__(self, num_nodes, dropout, in_dim, out_len, residual_channels, dilation_channels, skip_channels,
                 end_channels, kernel_size, blocks, layers):
        super(stcell, self).__init__()
        self.dropout = dropout
        self.blocks = blocks
        self.layers = layers
        self.filter_convs = nn.ModuleList()
        self.gate_convs = nn.ModuleList()
        self.residual_convs = nn.ModuleList()
        self.skip_convs = nn.ModuleList()
        self.ln = nn.ModuleList()
        self.gconv = nn.ModuleList()
        self.start_conv = nn.Conv2d(in_channels=in_dim,
                                    out_channels=residual_channels,
                                    kernel_size=(1,1))
        receptive_field = 1
        self.supports_len = 2

        for b in range(blocks):
            additional_scope = 1
            new_dilation = 1
            for i in range(layers):
                self.filter_convs.append(nn.Conv2d(in_channels=residual_channels,
                                                   out_channels=dilation_channels,
                                                   kernel_size=(1,kernel_size),dilation=new_dilation))

                self.gate_convs.append(nn.Conv2d(in_channels=residual_channels,
                                                 out_channels=dilation_channels,
                                                 kernel_size=(1, kernel_size), dilation=new_dilation))

                self.residual_convs.append(nn.Conv2d(in_channels=dilation_channels,
                                                     out_channels=residual_channels,
                                                     kernel_size=(1, 1)))

                self.skip_convs.append(nn.Conv2d(in_channels=dilation_channels,
                                                 out_channels=skip_channels,
                                                 kernel_size=(1, 1)))
                new_dilation *=2
                receptive_field += additional_scope
                additional_scope *= 2
                self.ln.append(nn.LayerNorm([residual_channels, num_nodes, (2 ** layers - 1) * blocks + 2 - receptive_field]))
                self.gconv.append(gcn(dilation_channels,residual_channels,dropout,support_len=self.supports_len))


        self.end_conv_b1 = nn.Conv2d(in_channels=skip_channels * blocks * layers,
                                  out_channels=end_channels,
                                  kernel_size=(1,1),
                                  bias=True)

        self.end_conv_b2 = nn.Conv2d(in_channels=end_channels,
                                    out_channels=out_len,
                                    kernel_size=(1,1),
                                    bias=True)

        self.end_conv_g1 = nn.Conv2d(in_channels=skip_channels* blocks * layers,
                                  out_channels=end_channels,
                                  kernel_size=(1,1),
                                  bias=True)

        self.end_conv_g2 = nn.Conv2d(in_channels=end_channels,
                                    out_channels=out_len,
                                    kernel_size=(1, 1),
                                    bias=True)

        self.receptive_field = receptive_field

    def forward(self, input, adp_g):
        in_len = input.size(3)
        if in_len<self.receptive_field:
            x = nn.functional.pad(input,(self.receptive_field-in_len,0,0,0))
        else:
            x = input
        x = self.start_conv(x)
        skip = 0

        for i in range(self.blocks * self.layers):
            res = x
            filter = self.filter_convs[i](x)
            filter = torch.tanh(filter)
            gate = self.gate_convs[i](x)
            gate = torch.sigmoid(gate)
            x = filter * gate

            s = x
            s = self.skip_convs[i](s)
            try:
                skip = torch.cat((s, skip[:, :, :,  -s.size(3):]), dim=1)
            except:
                skip = s

            x = self.gconv[i](x, adp_g)

            try:
                dense = dense[:, :, :, -x.size(3):]
            except:
                dense = 0
            dense = res[:, :, :, -x.size(3):] + dense

            gate = torch.sigmoid(x)
            x = x * gate + dense * (1 - gate)
            x = self.ln[i](x)

        param_b = F.relu(skip)
        param_b = F.relu(self.end_conv_b1(param_b))
        param_b = torch.sigmoid(self.end_conv_b2(param_b))

        param_g = F.relu(skip)
        param_g = F.relu(self.end_conv_g1(param_g))
        param_g = torch.sigmoid(self.end_conv_g2(param_g))

        return param_b, param_g


class SIRcell(nn.Module):
    def __init__(self):
        super(SIRcell, self).__init__()

    def forward(self, param_b: torch.Tensor, param_g: torch.Tensor, mob: torch.Tensor, SIR: torch.Tensor):
        if len(mob.shape) == 2:
            batch_size = SIR.shape[0]
            mob = mob.unsqueeze(0).expand(batch_size, -1, -1)
        num_node = SIR.shape[-2]
        S = SIR[..., [0]]
        I = SIR[..., [1]]
        R = SIR[..., [2]]
        pop = (S + I + R).expand(-1, num_node, num_node)
        propagtion = (mob/pop * I.expand(-1, num_node, num_node)).sum(1) +\
                     (mob/pop * I.expand(-1, num_node, num_node).transpose(1, 2)).sum(2)
        propagtion = propagtion.unsqueeze(2)

        I_new = param_b * propagtion
        R_t = I * param_g + R
        I_t = I + I_new - I * param_g
        S_t = S - I_new

        Ht_SIR = torch.cat((I_new, S_t, I_t, R_t), dim=-1)

        return Ht_SIR


class MepoGNN(BaseModel):
    """
    Meta-Population Graph Neural Network (MepoGNN)

    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features per node per timestep.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    glm_type : str, optional
        Type of graph learning model ('Dynamic', 'Adaptive'). Default: 'Dynamic'.
    adapt_graph : tensor, optional
        Initial tensor for adaptive adjacency matrix. Only needed if glm_type is 'Adaptive'. Default: None.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    residual_channels : int
        Number of channels in residual layers.
    dilation_channels : int
        Number of channels in dilation layers.
    skip_channels : int
        Number of channels in skip connection layers.
    end_channels : int
        Number of channels in the final convolution layers.
    kernel_size : int
        Kernel size for the convolution layers.
    blocks : int
        Number of blocks in the WaveNet architecture.
    layers : int
        Number of layers in each block.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output, num_nodes), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.
    """
    def __init__(self,
                 num_nodes, 
                 num_features, 
                 num_timesteps_input, 
                 num_timesteps_output, 
                 glm_type='Dynamic', 
                 adapt_graph=None, 
                 dropout=0.5, 
                 residual_channels=32,
                 dilation_channels=32, 
                 skip_channels=256, 
                 end_channels=512, 
                 kernel_size=2, 
                 blocks=2, 
                 layers=3, 
                 device = 'cpu'):
        super(MepoGNN, self).__init__()
        self.stcell = stcell(num_nodes, dropout, num_features, num_timesteps_output, residual_channels, dilation_channels,
                             skip_channels, end_channels, kernel_size, blocks, layers)
        self.SIRcell = SIRcell()
        self.out_dim = num_timesteps_output
        self.glm_type = glm_type
        self.device = device

        if self.glm_type == 'Adaptive':
            # To prevent parameter magnitude being too big
            log_g = torch.log(adapt_graph+1.0)
            self.max_log = log_g.max()
            # initialize g
            self.g_rescaled = nn.Parameter(log_g/self.max_log, requires_grad=True)

        elif self.glm_type == 'Dynamic':
            self.inc_init = nn.Parameter(torch.empty(num_timesteps_output, num_timesteps_input), requires_grad=True)
            nn.init.normal_(self.inc_init, 1, 0.01)
            self.od_scale_factor = 3
        else:
            raise NotImplementedError('Invalid graph type.')

    def forward(self, x, adj, states, dynamic_adj, max_od=1e6):
        """
        Parameters
        ----------
        x : torch.Tensor
            Input features tensor with shape (batch_size, num_timesteps_input, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_timesteps_output, num_nodes),
            representing the predicted values for each node over the specified output timesteps.
        """
        x_node = x
        od = dynamic_adj
        SIR = states

        x_node = x_node.transpose(1,3)
        if self.glm_type == 'Adaptive':
            mob = torch.exp(torch.relu(self.g_rescaled*self.max_log))
            g_adp = [mob / mob.sum(1, True), mob.T / mob.T.sum(1, True)]
            param_b, param_g = self.stcell(x_node, g_adp)
            outputs_SIR = []
            SIR = SIR[:, -1, ...]
            for i in range(self.out_dim):
                NSIR = self.SIRcell(param_b[:, i, ...], param_g[:, i, ...], mob, SIR)
                SIR = NSIR[..., 1:]
                outputs_SIR.append(NSIR[..., [0]])

        if self.glm_type == 'Dynamic':
            incidence = torch.softmax(self.inc_init, dim=1)
            mob = torch.einsum('kl,blnmc->bknmc', incidence, od).squeeze(-1)
            g = mob.mean(1)
            g_t = g.permute(0, 2, 1)
            g_dyn = [g / g.sum(2, True), g_t / g_t.sum(2, True)]
            param_b, param_g = self.stcell(x_node, g_dyn)
            outputs_SIR = []
            SIR = SIR[:, -1, ...]
            for i in range(self.out_dim):
                NSIR = self.SIRcell(param_b[:,i,...], param_g[:,i,...], mob[:,i,...]*max_od*self.od_scale_factor, SIR)
                SIR = NSIR[...,1:]
                outputs_SIR.append(NSIR[...,[0]])

        outputs = torch.stack(outputs_SIR, dim=1)

        return outputs.squeeze()
    
    def initialize(self):
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()



from . import (
            STGCN, 
            ATMGNN, 
            CNNRNN_Res, 
            STAN, 
            ColaGNN, 
            EpiColaGNN, 
            GraphLSTM, 
            MepoGNN, 
            EpiGNN, 
            DASTGN, 
            HierST, 
            DCRNN,
            GraphWaveNet,
            DMP,
            NetworkSIR,
)



import torch
import os
import matplotlib.pyplot as plt
from epilearn.models.SpatialTemporal.STGCN import STGCN
from epilearn.models.SpatialTemporal.ATMGNN import MPNN_LSTM, ATMGNN
from epilearn.models.SpatialTemporal.CNNRNN_Res import CNNRNN_Res
from epilearn.models.SpatialTemporal.STAN import STAN
from epilearn.models.SpatialTemporal.ColaGNN import COLAGNN

from epilearn.data import UniversalDataset
from epilearn.utils import utils

# initial settings
device = torch.device('cpu')
torch.manual_seed(7)

lookback = 13 # inputs size
horizon = 3 # predicts size

permute = False
target_feat_idx = None
target_idx = None

epochs = 50 # training epochs
batch_size = 50 # training batch size

# load toy dataset
dataset = UniversalDataset()
dataset.load_toy_dataset()

# preprocessing
features, mean, std = utils.normalize(dataset.x)
adj_norm = utils.normalize_adj(dataset.graph)

features = features.to(device)
adj_norm = adj_norm.to(device)

# prepare datasets
train_rate = 0.6 
val_rate = 0.2

split_line1 = int(features.shape[0] * train_rate)
split_line2 = int(features.shape[0] * (train_rate + val_rate))

train_original_data = features[:split_line1, :, :]
val_original_data = features[split_line1:split_line2, :, :]
test_original_data = features[split_line2:, :, :]

train_input, train_target = dataset.generate_dataset(X = train_original_data, Y = train_original_data[:, :, 0], lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)
val_input, val_target = dataset.generate_dataset(X = val_original_data, Y = val_original_data[:, :, 0], lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)
test_input, test_target = dataset.generate_dataset(X = test_original_data, Y = test_original_data[:, :, 0], lookback_window_size = lookback, horizon_size = horizon, permute = permute, feat_idx = target_feat_idx, target_idx = target_idx)

# prepare model

# model = STGCN(
#             num_nodes = adj_norm.shape[0],
#             num_features = train_input.shape[3],
#             num_timesteps_input = lookback,
#             num_timesteps_output = horizon
#             ).to(device = device)

# model = MPNN_LSTM(
#                 num_nodes = adj_norm.shape[0],
#                 num_features = train_input.shape[3],
#                 num_timesteps_input = lookback,
#                 num_timesteps_output = horizon,
#                 nhid = 4
#                 ).to(device = device)

# model = ATMGNN(
#                 num_nodes = adj_norm.shape[0],
#                 num_features = train_input.shape[3],
#                 num_timesteps_input = lookback,
#                 num_timesteps_output = horizon,
#                 nhid = 4
#                 ).to(device = device)

# model = CNNRNN_Res(
#                 num_nodes = adj_norm.shape[0],
#                 num_timesteps_input = lookback,
#                 num_timesteps_output = horizon,
#                 nhid = 4
#                 ).to(device = device)


# model = STAN(
#         num_nodes=adj_norm.shape[0],
#         num_features=train_input.shape[3],
#         num_timesteps_input=lookback,
#         num_timesteps_output=horizon,
#         gat_dim1=32, 
#         gat_dim2=32, 
#         gru_dim=32, 
#         num_heads=1,
#         device=device
#         ).to(device=device)

model = COLAGNN(
        num_nodes=adj_norm.shape[0],
        num_features=train_input.shape[3],
        num_timesteps_input=lookback,
        num_timesteps_output=horizon,
        nhid = 32,
        n_channels = 6,
        device = device,
        ).to(device=device)


# training
model.fit(
        train_input=train_input, 
        train_target=train_target, 
        graph=adj_norm, 
        val_input=val_input, 
        val_target=val_target, 
        verbose=True,
        batch_size=batch_size,
        epochs=epochs)


import torch.nn as nn
from .base import BaseModel
import torch.nn.init as init
import torch


class moving_avg(nn.Module):
    """
    Moving average block to highlight the trend of time series
    """

    def __init__(self, kernel_size, stride):
        super(moving_avg, self).__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(kernel_size=kernel_size, stride=stride, padding=0)

    def forward(self, x):
        # padding on the both ends of time series
        front = x[:, 0:1, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        end = x[:, -1:, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x

class series_decomp(nn.Module):
    """
    Series decomposition block
    """

    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean


class DlinearModel(BaseModel):
    """
    Dynamic Linear Model

    Parameters
    ----------
    num_features : int
        Number of features in each timestep of the input data.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.

    Attributes
    ----------
    decomposition : function
        Method to decompose the time series data into seasonal and trend components.
    Linear_Transform : torch.nn.Linear
        Linear transformation layer to project the decomposed input data to the output space.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output) representing the predicted values for the future timesteps.
        Each element corresponds to a predicted value for a specific future timestep. The output is averaged across the feature dimension to reduce it to a single predictive value per timestep.

    """
    def __init__(self, num_features, num_timesteps_input, num_timesteps_output):
        super(DlinearModel, self).__init__()
        self.num_features = num_features
        self.num_timesteps_input = num_timesteps_input
        self.num_timesteps_output = num_timesteps_output
        self.decomposition = series_decomp(kernel_size=3) 
        # Single Linear layer to handle all features
        self.Linear_Transform = nn.Linear(num_timesteps_input * num_features, num_timesteps_output * num_features)

    def forward(self, x):
        """
        Parameters
        ----------
        x : torch.Tensor
            The input tensor representing time series data for each batch. Expected shape is 
            (batch_size, num_timesteps_input, num_features), where `batch_size` is the number of samples in the batch,
            `num_timesteps_input` is the number of input time steps, and `num_features` represents features at each time step.

        Returns
        -------
        torch.Tensor
            The output of the model, a tensor of shape (batch_size, num_timesteps_output) that represents the predicted values 
            for the future time steps, reduced to a single predictive value per time step by averaging across the feature dimension.
        """
        # Decompose and process as before
        seasonal_init, trend_init = self.decomposition(x)
        batch_size = x.shape[0]
        
        # Flatten the input for linear transformation
        seasonal_flat = seasonal_init.view(batch_size, -1)
        trend_flat = trend_init.view(batch_size, -1)
        
        # Apply linear transformations
        seasonal_output_flat = self.Linear_Transform(seasonal_flat)
        trend_output_flat = self.Linear_Transform(trend_flat)
        
        # Reshape back and combine outputs
        seasonal_output = seasonal_output_flat.view(batch_size, self.num_timesteps_output, self.num_features)
        trend_output = trend_output_flat.view(batch_size, self.num_timesteps_output, self.num_features)
        output = seasonal_output + trend_output
        
        # Reduce feature dimension by averaging across features
        output = output.mean(dim=2)  # Average across the feature dimension
        # print(output.shape)  # Should now be [50, 3]
        return output

    def reset_parameters(self):
        # Here you reset parameters for each layer explicitly
        self.Linear_Transform.reset_parameters()
        # Add any other layers that need resetting
        # Note: Add here if decomposition or its components have resettable parameters

    def initialize(self):
        # This initializes parameters by resetting them for all children layers automatically
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()
  
